# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

import contextlib
import inspect
import os
import typing
from collections import defaultdict

try:
    from opentelemetry import trace as trace_api
    from opentelemetry.sdk.trace import ReadableSpan, TracerProvider
    from opentelemetry.sdk.trace.export import SimpleSpanProcessor
    from opentelemetry.sdk.trace.export.in_memory_span_exporter import (
        InMemorySpanExporter,
    )
    from opentelemetry.sdk.resources import (
        SERVICE_INSTANCE_ID,
        SERVICE_NAME,
        Resource,
    )

    opentelemetry_packages_available = True
except ImportError:
    opentelemetry_packages_available = False

import pytest

from tango import DevState, DeviceProxy, GreenMode, constants
from tango.asyncio import DeviceProxy as AsyncioDeviceProxy
from tango.server import Device, attribute, command
from tango.utils import (
    _telemetry_active,
    get_telemetry_tracer_provider_factory,
    set_telemetry_tracer_provider_factory,
)
from tango.test_utils import DeviceTestContext


@pytest.fixture()
def exporters():
    """Switch to in-memory exporters for Python telemetry"""
    in_mem_exporters: dict[str, InMemorySpanExporter] = {}

    def in_memory_tracer_provider(
        service_name,
        service_instance_id=None,
        extra_resource_attributes=None,
    ):
        resource_attributes = {SERVICE_NAME: service_name}
        if service_instance_id:
            resource_attributes[SERVICE_INSTANCE_ID] = service_instance_id
        tracer_provider = TracerProvider(resource=Resource.create(resource_attributes))
        exporter = InMemorySpanExporter()
        in_mem_exporters[service_name] = exporter
        processor = SimpleSpanProcessor(exporter)
        tracer_provider.add_span_processor(processor)
        return tracer_provider

    old_factory = get_telemetry_tracer_provider_factory()
    set_telemetry_tracer_provider_factory(in_memory_tracer_provider)
    yield in_mem_exporters
    set_telemetry_tracer_provider_factory(old_factory)


class CapturedTelemetry:
    def __init__(self, exporters):
        self._exporters = exporters
        self._device_class_name = ""
        self._client_spans: dict[str, list["ReadableSpan"]] = defaultdict(list)
        self._device_spans: dict[str, list["ReadableSpan"]] = defaultdict(list)

    def set_device_class_name(self, name):
        self._device_class_name = name

    @property
    def client_startup_spans(self) -> list["ReadableSpan"]:
        return self._client_spans["startup"]

    @property
    def device_startup_spans(self) -> list["ReadableSpan"]:
        return self._device_spans["startup"]

    @property
    def client_running_spans(self) -> list["ReadableSpan"]:
        return self._client_spans["running"]

    @property
    def device_running_spans(self) -> list["ReadableSpan"]:
        return self._device_spans["running"]

    @property
    def client_shutdown_spans(self) -> list["ReadableSpan"]:
        return self._client_spans["shutdown"]

    @property
    def device_shutdown_spans(self) -> list["ReadableSpan"]:
        return self._device_spans["shutdown"]

    def startup_done(self):
        self._stage_done("startup")

    def running_done(self):
        self._stage_done("running")

    def shutdown_done(self):
        self._stage_done("shutdown")

    def ignore_recent_spans(self):
        self._stage_done("ignore")

    def _stage_done(self, stage):
        client = self._exporters.get("pytango.client")
        if client:
            client_spans = client.get_finished_spans()
            client.clear()
        else:
            client_spans = []
            print(f"No client for spans. {self._exporters=}")
        device = self._exporters.get(self._device_class_name)
        if device:
            device_spans = device.get_finished_spans()
            device.clear()
        else:
            device_spans = []
            print("No device for spans")
        self._client_spans[stage] = client_spans
        self._device_spans[stage] = device_spans


def print_json(spans):
    """Utility that is useful when debugging tests"""
    print(f"===== Printing {len(spans)} spans... ========")
    for span in spans:
        print(span.to_json())
        print("-----------------------")
    print(f"===== Done printing {len(spans)} spans ========")


@pytest.fixture()
def simple_device(server_green_mode):
    # Note: Telemetry spans are not emitted for BaseDevice methods by default,
    # so we override init_device, delete_device and dev_state in our test device.
    # This lets us verify that user methods will generate spans.

    if server_green_mode == GreenMode.Asyncio:

        class TestDevice(Device):
            green_mode = server_green_mode

            async def init_device(self):
                await super().init_device()

            async def delete_device(self):
                await super().delete_device()

            async def dev_state(self):
                return DevState.RUNNING

            @attribute
            async def lineno_attribute(self) -> int:
                return inspect.currentframe().f_lineno - 2

            @command
            async def lineno_command(self) -> int:
                return inspect.currentframe().f_lineno - 2

    else:

        class TestDevice(Device):
            green_mode = server_green_mode

            def init_device(self):
                super().init_device()

            def delete_device(self):
                super().delete_device()

            def dev_state(self):
                return DevState.RUNNING

            @attribute
            def lineno_attribute(self) -> int:
                return inspect.currentframe().f_lineno - 2

            @command
            def lineno_command(self) -> int:
                return inspect.currentframe().f_lineno - 2

    return TestDevice


@contextlib.contextmanager
def span_recording_device_test_context(
    telemetry: CapturedTelemetry, device_class: typing.Type[Device], **kwargs
) -> typing.Generator[DeviceProxy, None, None]:
    """Context manager that records the telemetry spans around DeviceTestContext.

    This lets us capture the spans created at various stages:
      - on device startup
      - while the device is running (if test function accesses device via proxy)
      - on device shutdown
    """
    telemetry.set_device_class_name(device_class.__name__)
    context = DeviceTestContext(device_class, **kwargs)
    context.start()
    try:
        telemetry.startup_done()
        yield context.device
        telemetry.running_done()
    finally:
        context.stop()
        context.join()
    telemetry.shutdown_done()


def test_telemetry_available_constant_exists():
    assert isinstance(constants.TELEMETRY_SUPPORTED, bool)


@pytest.mark.skipif(not _telemetry_active, reason="Telemetry not active")
def test_telemetry_packages_available_if_telemetry_active():
    assert opentelemetry_packages_available


@pytest.mark.skipif(not _telemetry_active, reason="Telemetry not active")
def test_init_device_and_basic_span_details(exporters, simple_device):
    telemetry = CapturedTelemetry(exporters)
    with span_recording_device_test_context(telemetry, simple_device):
        pass

    assert len(telemetry.client_startup_spans) > 0
    assert len(telemetry.device_startup_spans) > 0

    client_span = telemetry.client_startup_spans[0]
    assert client_span.name == "span_recording_device_test_context"
    assert client_span.resource.attributes[SERVICE_NAME] == "pytango.client"
    assert client_span.attributes["code.filepath"] == __file__
    assert "code.lineno" in client_span.attributes
    assert "thread.id" in client_span.attributes
    assert "thread.name" in client_span.attributes

    device_span = telemetry.device_startup_spans[0]
    assert device_span.name == "simple_device.<locals>.TestDevice.init_device"
    assert device_span.resource.attributes[SERVICE_NAME] == "TestDevice"
    assert (
        device_span.resource.attributes[SERVICE_INSTANCE_ID] == "test/nodb/testdevice"
    )
    assert device_span.attributes["code.filepath"] == __file__
    assert "code.lineno" in device_span.attributes
    assert "thread.id" in device_span.attributes
    assert "thread.name" in device_span.attributes


@pytest.mark.skipif(not _telemetry_active, reason="Telemetry not active")
def test_delete_device(exporters, simple_device):
    telemetry = CapturedTelemetry(exporters)
    with span_recording_device_test_context(telemetry, simple_device):
        pass

    assert len(telemetry.client_shutdown_spans) == 1
    assert len(telemetry.device_shutdown_spans) == 1

    client_span = telemetry.client_shutdown_spans[0]
    assert client_span.name == "span_recording_device_test_context"

    device_span = telemetry.device_shutdown_spans[0]
    assert device_span.name == "simple_device.<locals>.TestDevice.delete_device"


@pytest.mark.skipif(not _telemetry_active, reason="Telemetry not active")
def test_state(exporters, simple_device):
    telemetry = CapturedTelemetry(exporters)
    with span_recording_device_test_context(telemetry, simple_device) as proxy:
        proxy.State()
        state_lineno = inspect.currentframe().f_lineno - 1

    assert_single_client_and_device_running_span_and_share_trace_id(telemetry)

    client_span = telemetry.client_running_spans[0]
    assert client_span.name == "test_state"
    assert client_span.attributes["code.lineno"] == state_lineno

    device_span = telemetry.device_running_spans[0]
    assert device_span.name == "simple_device.<locals>.TestDevice.dev_state"


@pytest.mark.skipif(not _telemetry_active, reason="Telemetry not active")
def test_static_command(exporters, simple_device):
    telemetry = CapturedTelemetry(exporters)
    with span_recording_device_test_context(telemetry, simple_device) as proxy:
        device_lineno = proxy.lineno_command()
        client_lineno = inspect.currentframe().f_lineno - 1

    assert_single_client_and_device_running_span_and_share_trace_id(telemetry)

    client_span = telemetry.client_running_spans[0]
    assert client_span.name == "test_static_command"
    assert client_span.attributes["code.lineno"] == client_lineno

    device_span = telemetry.device_running_spans[0]
    assert device_span.name == "simple_device.<locals>.TestDevice.lineno_command"
    assert device_span.attributes["code.lineno"] == device_lineno


@pytest.mark.skipif(not _telemetry_active, reason="Telemetry not active")
def test_static_attribute(exporters, simple_device, green_mode_device_proxy):
    telemetry = CapturedTelemetry(exporters)
    with span_recording_device_test_context(telemetry, simple_device) as proxy:
        gm_proxy = green_mode_device_proxy(proxy.dev_name())  # emits some spans
        telemetry.ignore_recent_spans()
        device_lineno = gm_proxy.read_attribute("lineno_attribute", wait=True).value
        client_lineno = inspect.currentframe().f_lineno - 1

    assert_single_client_and_device_running_span_and_share_trace_id(telemetry)

    client_span = telemetry.client_running_spans[0]
    assert client_span.name == "test_static_attribute"
    assert client_span.attributes["code.lineno"] == client_lineno

    device_span = telemetry.device_running_spans[0]
    assert device_span.name == "simple_device.<locals>.TestDevice.lineno_attribute"
    assert device_span.attributes["code.lineno"] == device_lineno


def assert_single_client_and_device_running_span_and_share_trace_id(telemetry):
    assert len(telemetry.client_running_spans) == 1
    assert len(telemetry.device_running_spans) == 1

    client_id = telemetry.client_running_spans[0].context.trace_id
    device_id = telemetry.device_running_spans[0].context.trace_id
    assert client_id == device_id


@pytest.mark.skipif(not _telemetry_active, reason="Telemetry not active")
@pytest.mark.asyncio
async def test_static_attribute_asyncio(exporters, simple_device):
    telemetry = CapturedTelemetry(exporters)
    with span_recording_device_test_context(telemetry, simple_device) as proxy:
        aproxy = await AsyncioDeviceProxy(proxy.dev_name())  # emits some spans
        telemetry.ignore_recent_spans()
        _ = await aproxy.lineno_attribute
        client_lineno = inspect.currentframe().f_lineno - 1

    assert_single_client_and_device_running_span_and_share_trace_id(telemetry)

    client_span = telemetry.client_running_spans[0]
    assert client_span.name == "test_static_attribute_asyncio"
    assert client_span.attributes["code.lineno"] == client_lineno


@pytest.mark.skipif(not _telemetry_active, reason="Telemetry not active")
def test_client_ident_included_for_device(exporters, simple_device):
    telemetry = CapturedTelemetry(exporters)
    with span_recording_device_test_context(telemetry, simple_device) as proxy:
        proxy.State()

    device_span = telemetry.device_running_spans[0]
    assert "collocated" in device_span.attributes["tango.client_ident.location"]
    assert device_span.attributes["tango.client_ident.pid"] == os.getpid()
    assert device_span.attributes["tango.client_ident.lang"].startswith("CPP")


@pytest.mark.skipif(not _telemetry_active, reason="Telemetry not active")
def test_user_span_traceid_propagates_to_tango(
    exporters, simple_device, green_mode_device_proxy
):
    factory = get_telemetry_tracer_provider_factory()
    user_provider = factory("user")
    user_tracer = trace_api.get_tracer("user.tracer", tracer_provider=user_provider)

    telemetry = CapturedTelemetry(exporters)
    with user_tracer.start_as_current_span("user.span"):
        with span_recording_device_test_context(telemetry, simple_device) as proxy:
            gm_proxy = green_mode_device_proxy(proxy.dev_name())  # emits some spans
            telemetry.ignore_recent_spans()
            _ = gm_proxy.command_inout("State", wait=True)

    user_spans = exporters["user"].get_finished_spans()
    assert len(user_spans) == 1
    assert len(telemetry.client_running_spans) == 1
    assert len(telemetry.device_running_spans) == 1

    user_trace_id = user_spans[0].context.trace_id
    client_trace_id = telemetry.client_running_spans[0].context.trace_id
    device_trace_id = telemetry.device_running_spans[0].context.trace_id
    assert client_trace_id == user_trace_id
    assert device_trace_id == user_trace_id


@pytest.mark.skipif(not _telemetry_active, reason="Telemetry not active")
def test_base_device_kernel_tracing_disabled_by_default(exporters):
    telemetry = CapturedTelemetry(exporters)
    with span_recording_device_test_context(telemetry, Device):
        pass

    assert len(telemetry.device_startup_spans) == 0
    assert len(telemetry.device_shutdown_spans) == 0


@pytest.mark.skipif(not _telemetry_active, reason="Telemetry not active")
def test_base_device_traces_if_kernel_tracing_enabled(exporters, simple_device):
    class TestDevice(simple_device):
        def create_telemetry_tracer_provider(self, *args, **kwargs):
            # we override create_telemetry_tracer_provider because it gets called
            # just before init_device
            self.set_kernel_tracing_enabled(True)
            return super().create_telemetry_tracer_provider(*args, **kwargs)

    telemetry = CapturedTelemetry(exporters)
    with span_recording_device_test_context(telemetry, TestDevice):
        pass

    startup_spans = telemetry.device_startup_spans
    shutdown_spans = telemetry.device_shutdown_spans
    assert len(startup_spans) == 3
    assert startup_spans[0].name == "BaseDevice.init_device"
    assert startup_spans[1].name == "simple_device.<locals>.TestDevice.init_device"
    assert startup_spans[2].name == "BaseDevice.server_init_hook"
    assert len(shutdown_spans) == 2
    assert shutdown_spans[0].name == "BaseDevice.delete_device"
    assert shutdown_spans[1].name == "simple_device.<locals>.TestDevice.delete_device"


@pytest.mark.skipif(not _telemetry_active, reason="Telemetry not active")
def test_no_device_traces_if_device_tracing_disabled(exporters, simple_device):
    class TestDevice(simple_device):
        def create_telemetry_tracer_provider(self, *args, **kwargs):
            # we override create_telemetry_tracer_provider because it gets called
            # just before init_device
            self.set_telemetry_enabled(False)
            return super().create_telemetry_tracer_provider(*args, **kwargs)

    telemetry = CapturedTelemetry(exporters)
    with span_recording_device_test_context(telemetry, TestDevice) as proxy:
        _ = proxy.lineno_attribute

    assert len(telemetry.device_startup_spans) == 0
    assert len(telemetry.device_running_spans) == 0
    assert len(telemetry.device_shutdown_spans) == 0



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
import time
import numpy as np
from functools import wraps

try:
    import numpy.typing as npt
except ImportError:
    npt = None

import pytest

from tango import (
    DevFailed,
    DevState,
    GreenMode,
    PyTangoUserWarning,  # noqa
    DispLevel,
    DevVarDoubleArray,
    DevLong64,
    DevVoid,
    CommandInfoList,
)
from tango.server import Device
from tango.server import command
from tango.pyutil import TimedCmdData
from tango.test_utils import DeviceTestContext
from tango.test_utils import (
    assert_close,
    general_decorator,
    general_asyncio_decorator,
    convert_dtype_to_typing_hint,
)


def test_static_commands(command_typed_values, server_green_mode):
    dtype, values, expected = command_typed_values

    if dtype == (bool,):
        pytest.xfail("Not supported for some reasons")

    if server_green_mode == GreenMode.Asyncio:

        class TestDevice(Device):
            green_mode = server_green_mode
            _last_value = None

            @command()
            async def void_cmd(self):
                pass

            @command(dtype_in=dtype)
            async def no_return_cmd(self, arg):
                self._last_value = arg

            @command(dtype_out=dtype)
            async def no_input_cmd(self):
                return self._last_value

            @command(dtype_in=dtype, dtype_out=dtype)
            async def identity(self, arg):
                return arg

    else:

        class TestDevice(Device):
            green_mode = server_green_mode
            _last_value = None

            @command()
            def void_cmd(self):
                pass

            @command(dtype_in=dtype)
            def no_return_cmd(self, arg):
                self._last_value = arg

            @command(dtype_out=dtype)
            def no_input_cmd(self):
                return self._last_value

            @command(dtype_in=dtype, dtype_out=dtype)
            def identity(self, arg):
                return arg

    with DeviceTestContext(TestDevice) as proxy:
        for value in values:
            # just check, that we can execute it
            proxy.void_cmd()
            proxy.no_return_cmd(value)
            assert_close(proxy.no_input_cmd(), expected(value))
            assert_close(proxy.identity(value), expected(value))


def test_static_commands_with_typing(command_typed_values):
    dtype, values, expected = command_typed_values
    tuple_hint, list_hint, _, _ = convert_dtype_to_typing_hint(dtype)

    if dtype == (bool,):
        pytest.xfail("Not supported for some reasons")

    class TestDevice(Device):
        _last_arg = None

        @command()
        def command_tuple_hint(self, arg: tuple_hint) -> tuple_hint:
            return arg

        @general_decorator
        @command()
        def command_with_decorator_before(self, arg: tuple_hint) -> tuple_hint:
            return arg

        @command()
        @general_decorator
        def command_with_decorator_after(self, arg: tuple_hint) -> tuple_hint:
            return arg

        @command()
        def command_list_hint(self, arg: list_hint) -> list_hint:
            return arg

        @command(dtype_in=dtype, dtype_out=dtype)
        def command_user_type_has_priority(self, arg: dict) -> dict:
            return arg

        @command(dtype_in=dtype)
        def command_in_type_dtype_out_type_hint(self, arg) -> tuple_hint:
            return arg

        @command(dtype_out=dtype)
        def command_in_type_hint_out_type_dtype(self, arg: tuple_hint):
            return arg

        @command()
        def command_only_in_type_hint(self, arg: tuple_hint):
            self._last_arg = arg

        @command()
        def command_only_out_type_hint(self) -> tuple_hint:
            return self._last_arg

        @command()
        def command_arg_list_in_only_out_type_hint(self, *args) -> tuple_hint:
            return self._last_arg

        @command()
        def reset(self):
            self._last_arg = None

    with DeviceTestContext(TestDevice) as proxy:
        for value in values:
            assert_close(proxy.command_tuple_hint(value), expected(value))
            assert_close(proxy.command_with_decorator_before(value), expected(value))
            assert_close(proxy.command_with_decorator_after(value), expected(value))
            assert_close(proxy.command_list_hint(value), expected(value))
            assert_close(proxy.command_user_type_has_priority(value), expected(value))
            assert_close(
                proxy.command_in_type_dtype_out_type_hint(value), expected(value)
            )
            proxy.command_only_in_type_hint(value)
            assert_close(proxy.command_only_out_type_hint(), expected(value))
            assert_close(
                proxy.command_arg_list_in_only_out_type_hint(), expected(value)
            )
            proxy.reset()


def test_devstate_command_with_typing():
    class TestDevice(Device):
        @command
        def arbitrary_devstate_command(self, arg_in: DevState) -> DevState:
            return arg_in

    with DeviceTestContext(TestDevice) as proxy:
        assert proxy.arbitrary_devstate_command(DevState.MOVING) == DevState.MOVING


def test_command_self_typed_with_not_defined_name():
    class TestDevice(Device):
        @command
        def identity(self: "TestDevice", arg_in: int) -> int:
            return arg_in

        def dynamic_identity(self: "TestDevice", arg_in: int) -> int:
            return arg_in

        @command()
        def add_dyn_cmd(self: "TestDevice"):
            cmd = command(f=self.dynamic_identity)
            self.add_command(cmd)

    with DeviceTestContext(TestDevice) as proxy:
        proxy.add_dyn_cmd()
        assert 1 == proxy.identity(1)
        assert 1 == proxy.dynamic_identity(1)


def test_decorated_command(server_green_mode):
    if server_green_mode == GreenMode.Asyncio:

        class TestDevice(Device):
            green_mode = server_green_mode
            is_allowed = None

            @command(dtype_in=int, dtype_out=int)
            @general_asyncio_decorator()
            async def identity(self, arg):
                return arg

            @general_asyncio_decorator
            async def is_identity_allowed(self):
                return self.is_allowed

            @command(dtype_in=bool)
            async def make_allowed(self, yesno):
                self.is_allowed = yesno

    else:

        class TestDevice(Device):
            green_mode = server_green_mode
            is_allowed = None

            @command(dtype_in=int, dtype_out=int)
            @general_decorator()
            def identity(self, arg):
                return arg

            @general_decorator
            def is_identity_allowed(self):
                return self.is_allowed

            @command(dtype_in=bool)
            def make_allowed(self, yesno):
                self.is_allowed = yesno

    with DeviceTestContext(TestDevice) as proxy:
        proxy.make_allowed(True)
        assert_close(proxy.identity(123), 123)

        proxy.make_allowed(False)
        with pytest.raises(DevFailed):
            proxy.identity(1)


def test_command_isallowed(server_green_mode):
    is_allowed = None

    def sync_allowed(device):
        assert isinstance(device, TestDevice)
        return is_allowed

    async def async_allowed(device):
        assert isinstance(device, TestDevice)
        return is_allowed

    class IsAllowedCallableClass:
        def __init__(self):
            self._is_allowed = None

        def __call__(self, device):
            assert isinstance(device, TestDevice)
            return self._is_allowed

        def make_allowed(self, yesno):
            self._is_allowed = yesno

    is_allowed_callable_class = IsAllowedCallableClass()

    class AsyncIsAllowedCallableClass(IsAllowedCallableClass):
        async def __call__(self, device):
            assert isinstance(device, TestDevice)
            return self._is_allowed

    async_is_allowed_callable_class = AsyncIsAllowedCallableClass()

    if server_green_mode == GreenMode.Asyncio:

        class TestDevice(Device):
            green_mode = server_green_mode

            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self._is_allowed = True

            @command(dtype_in=int, dtype_out=int)
            async def identity(self, arg):
                return arg

            @command(dtype_in=int, dtype_out=int, fisallowed="is_identity_allowed")
            async def identity_kwarg_string(self, arg):
                return arg

            @command(
                dtype_in=int,
                dtype_out=int,
                fisallowed=async_allowed,
            )
            async def identity_kwarg_callable(self, arg):
                return arg

            @command(
                dtype_in=int, dtype_out=int, fisallowed=async_is_allowed_callable_class
            )
            async def identity_kwarg_callable_class(self, arg):
                return arg

            @command(dtype_in=int, dtype_out=int)
            async def identity_always_allowed(self, arg):
                return arg

            @command(dtype_in=bool)
            async def make_allowed(self, yesno):
                self._is_allowed = yesno

            async def is_identity_allowed(self):
                return self._is_allowed

    else:

        class TestDevice(Device):
            green_mode = server_green_mode

            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self._is_allowed = True

            @command(dtype_in=int, dtype_out=int)
            def identity(self, arg):
                return arg

            @command(dtype_in=int, dtype_out=int, fisallowed="is_identity_allowed")
            def identity_kwarg_string(self, arg):
                return arg

            @command(dtype_in=int, dtype_out=int, fisallowed=sync_allowed)
            def identity_kwarg_callable(self, arg):
                return arg

            @command(dtype_in=int, dtype_out=int, fisallowed=is_allowed_callable_class)
            def identity_kwarg_callable_class(self, arg):
                return arg

            @command(dtype_in=int, dtype_out=int)
            def identity_always_allowed(self, arg):
                return arg

            @command(dtype_in=bool)
            def make_allowed(self, yesno):
                self._is_allowed = yesno

            def is_identity_allowed(self):
                return self._is_allowed

    with DeviceTestContext(TestDevice) as proxy:
        proxy.make_allowed(True)
        is_allowed_callable_class.make_allowed(True)
        async_is_allowed_callable_class.make_allowed(True)
        is_allowed = True

        assert_close(proxy.identity(1), 1)
        assert_close(proxy.identity_kwarg_string(1), 1)
        assert_close(proxy.identity_kwarg_callable(1), 1)
        assert_close(proxy.identity_kwarg_callable_class(1), 1)
        assert_close(proxy.identity_always_allowed(1), 1)

        proxy.make_allowed(False)
        is_allowed_callable_class.make_allowed(False)
        async_is_allowed_callable_class.make_allowed(False)
        is_allowed = False

        with pytest.raises(DevFailed):
            proxy.identity(1)

        with pytest.raises(DevFailed):
            proxy.identity_kwarg_string(1)

        with pytest.raises(DevFailed):
            proxy.identity_kwarg_callable(1)

        with pytest.raises(DevFailed):
            proxy.identity_kwarg_callable_class(1)

        assert_close(proxy.identity_always_allowed(1), 1)


@pytest.mark.parametrize("device_command_level", [True, False])
def test_dynamic_command(device_command_level, server_green_mode):
    is_allowed = None

    def sync_allowed(device):
        assert isinstance(device, TestDevice)
        return is_allowed

    async def async_allowed(device):
        assert isinstance(device, TestDevice)
        return is_allowed

    class IsAllowedCallable:
        def __init__(self):
            self._is_allowed = None

        def __call__(self, device):
            assert isinstance(device, TestDevice)
            return self._is_allowed

        def make_allowed(self, yesno):
            self._is_allowed = yesno

    class AsyncIsAllowedCallable(IsAllowedCallable):
        async def __call__(self, device):
            assert isinstance(device, TestDevice)
            return self._is_allowed

    is_allowed_callable_class = IsAllowedCallable()
    async_is_allowed_callable_class = AsyncIsAllowedCallable()

    class BaseTestDevice(Device):
        green_mode = server_green_mode
        _last_in = None

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._is_allowed = True

        def _add_dyn_cmd(self):
            cmd = command(f=self.void_cmd)
            self.add_command(cmd, device_command_level)

            cmd = command(f=self.no_return_cmd, dtype_in=int)
            self.add_command(cmd, device_command_level)

            cmd = command(f=self.no_input_cmd, dtype_out=int)
            self.add_command(cmd, device_command_level)

            cmd = command(f=self.no_input_args_list_cmd, dtype_out=int)
            self.add_command(cmd, device_command_level)

            cmd = command(f=self.identity, dtype_in=int, dtype_out=int)
            self.add_command(cmd, device_command_level)

            cmd = command(
                f=self.identity_kwarg_string,
                dtype_in=int,
                dtype_out=int,
                fisallowed="is_identity_allowed",
            )
            self.add_command(cmd, device_command_level)

            cmd = command(
                f=self.identity_kwarg_callable,
                dtype_in=int,
                dtype_out=int,
                fisallowed=self.is_identity_allowed,
            )
            self.add_command(cmd, device_command_level)

            cmd = command(
                f=self.identity_kwarg_callable_outside_class,
                dtype_in=int,
                dtype_out=int,
                fisallowed=(
                    sync_allowed
                    if server_green_mode != GreenMode.Asyncio
                    else async_allowed
                ),
            )
            self.add_command(cmd, device_command_level)

            cmd = command(
                f=self.identity_kwarg_callable_class,
                dtype_in=int,
                dtype_out=int,
                fisallowed=(
                    is_allowed_callable_class
                    if server_green_mode != GreenMode.Asyncio
                    else async_is_allowed_callable_class
                ),
            )
            self.add_command(cmd, device_command_level)

            cmd = command(f=self.identity_always_allowed, dtype_in=int, dtype_out=int)
            self.add_command(cmd, device_command_level)

    if server_green_mode == GreenMode.Asyncio:

        class TestDevice(BaseTestDevice):
            async def void_cmd(self):
                pass

            async def no_return_cmd(self, arg):
                self._last_in = arg

            async def no_input_cmd(self):
                return self._last_in

            async def no_input_args_list_cmd(self, *args):
                return self._last_in

            async def identity(self, arg):
                return arg

            async def identity_kwarg_string(self, arg):
                return arg

            async def identity_kwarg_callable(self, arg):
                return arg

            async def identity_kwarg_callable_outside_class(self, arg):
                return arg

            async def identity_kwarg_callable_class(self, arg):
                return arg

            async def identity_always_allowed(self, arg):
                return arg

            @command()
            async def add_dyn_cmd(self):
                self._add_dyn_cmd()

            @command(dtype_in=bool)
            async def make_allowed(self, yesno):
                self._is_allowed = yesno

            async def is_identity_allowed(self):
                return self._is_allowed

    else:

        class TestDevice(BaseTestDevice):
            def void_cmd(self):
                pass

            def no_return_cmd(self, arg):
                self._last_in = arg

            def no_input_cmd(self):
                return self._last_in

            def no_input_args_list_cmd(self, *args):
                return self._last_in

            def identity(self, arg):
                return arg

            def identity_kwarg_string(self, arg):
                return arg

            def identity_kwarg_callable(self, arg):
                return arg

            def identity_kwarg_callable_outside_class(self, arg):
                return arg

            def identity_kwarg_callable_class(self, arg):
                return arg

            def identity_always_allowed(self, arg):
                return arg

            @command()
            def add_dyn_cmd(self):
                self._add_dyn_cmd()

            @command(dtype_in=bool)
            def make_allowed(self, yesno):
                self._is_allowed = yesno

            def is_identity_allowed(self):
                return self._is_allowed

    with DeviceTestContext(TestDevice) as proxy:
        proxy.add_dyn_cmd()

        proxy.make_allowed(True)
        is_allowed_callable_class.make_allowed(True)
        async_is_allowed_callable_class.make_allowed(True)
        is_allowed = True

        # check that cmd exist and can be executed
        proxy.void_cmd()

        proxy.no_return_cmd(1)
        assert_close(proxy.no_input_cmd(), 1)

        assert_close(proxy.identity(1), 1)
        assert_close(proxy.identity_kwarg_string(1), 1)
        assert_close(proxy.identity_kwarg_callable(1), 1)
        assert_close(proxy.identity_kwarg_callable_outside_class(1), 1)
        assert_close(proxy.identity_kwarg_callable_class(1), 1)
        assert_close(proxy.identity_always_allowed(1), 1)

        proxy.make_allowed(False)
        is_allowed_callable_class.make_allowed(False)
        async_is_allowed_callable_class.make_allowed(False)
        is_allowed = False

        with pytest.raises(DevFailed):
            proxy.identity(1)

        with pytest.raises(DevFailed):
            proxy.identity_kwarg_string(1)

        with pytest.raises(DevFailed):
            proxy.identity_kwarg_callable(1)

        with pytest.raises(DevFailed):
            proxy.identity_kwarg_callable_outside_class(1)

        with pytest.raises(DevFailed):
            proxy.identity_kwarg_callable_class(1)

        assert_close(proxy.identity_always_allowed(1), 1)


def test_identity_dynamic_command_with_typing(command_typed_values):
    dtype, values, expected = command_typed_values
    tuple_hint, list_hint, _, _ = convert_dtype_to_typing_hint(dtype)

    if dtype == (bool,):
        pytest.xfail("Not supported for some reasons")

    class TestDevice(Device):
        def command_tuple_hint(self, arg: tuple_hint) -> tuple_hint:
            return arg

        def command_list_hint(self, arg: list_hint) -> list_hint:
            return arg

        def command_user_type_has_priority(self, arg: dict) -> dict:
            return arg

        def command_args_list_tuple_hint(self, *args) -> tuple_hint:
            return values[0]

        @command()
        def add_dyn_cmd(self):
            cmd = command(f=self.command_tuple_hint)
            self.add_command(cmd)

            cmd = command(f=self.command_list_hint)
            self.add_command(cmd)

            cmd = command(
                f=self.command_user_type_has_priority, dtype_in=dtype, dtype_out=dtype
            )
            self.add_command(cmd)

            cmd = command(f=self.command_args_list_tuple_hint)
            self.add_command(cmd)

    with DeviceTestContext(TestDevice) as proxy:
        proxy.add_dyn_cmd()
        for value in values:
            assert_close(proxy.command_tuple_hint(value), expected(value))
            assert_close(proxy.command_list_hint(value), expected(value))
            assert_close(proxy.command_user_type_has_priority(value), expected(value))
        assert_close(proxy.command_args_list_tuple_hint(), expected(values[0]))


if npt:

    def test_identity_commands_with_numpy_typing(command_numpy_typed_values):
        type_hint, dformat, value, expected = command_numpy_typed_values
        if type_hint == np.uint8:
            pytest.xfail("Does not work for some reason")

        class TestDevice(Device):
            def identity_dynamic_command(self, arg: type_hint) -> type_hint:
                return arg

            @command
            def identity_static_command(self, arg: type_hint) -> type_hint:
                return arg

            @command()
            def add_dyn_cmd(self):
                cmd = command(f=self.identity_dynamic_command)
                self.add_command(cmd)

        with DeviceTestContext(TestDevice) as proxy:
            proxy.add_dyn_cmd()
            assert_close(proxy.identity_static_command(value), expected(value))
            assert_close(proxy.identity_dynamic_command(value), expected(value))


def test_polled_command():
    dct = {"Polling1": 100, "Polling2": 100000, "Polling3": 500}

    class TestDevice(Device):

        @command(polling_period=dct["Polling1"])
        def Polling1(self):
            pass

        @command(polling_period=dct["Polling2"])
        def Polling2(self):
            pass

        @command(polling_period=dct["Polling3"])
        def Polling3(self):
            pass

    with DeviceTestContext(TestDevice) as proxy:
        ans = proxy.polling_status()

    for info in ans:
        lines = info.split("\n")
        comm = lines[0].split("= ")[1]
        period = int(lines[1].split("= ")[1])
        assert dct[comm] == period


def test_wrong_command_result():
    class TestDevice(Device):

        @command(dtype_out=str)
        def cmd_str_err(self):
            return 1.2345

        @command(dtype_out=int)
        def cmd_int_err(self):
            return "bla"

        @command(dtype_out=[str])
        def cmd_str_list_err(self):
            return ["hello", 55]

    with DeviceTestContext(TestDevice) as proxy:
        with pytest.raises(DevFailed):
            proxy.cmd_str_err()
        with pytest.raises(DevFailed):
            proxy.cmd_int_err()
        with pytest.raises(DevFailed):
            proxy.cmd_str_list_err()


def test_command_info():
    class TestDevice(Device):

        @command(
            doc_in="identity_scalar doc_in",
            doc_out="identity_scalar doc_out",
            display_level=DispLevel.OPERATOR,
        )
        def identity_scalar(self, val: int) -> int:
            return val

        @command(
            doc_in="identity_spectrum doc_in",
            doc_out="identity_spectrum doc_out",
            display_level=DispLevel.EXPERT,
        )
        def identity_spectrum(self, val: list[float]) -> list[float]:
            return val

    with DeviceTestContext(TestDevice) as proxy:
        info = proxy.get_command_config("identity_scalar")
        assert info.cmd_name == "identity_scalar"
        assert info.disp_level == DispLevel.OPERATOR
        assert info.in_type == DevLong64
        assert info.in_type_desc == "identity_scalar doc_in"
        assert info.out_type == DevLong64
        assert info.out_type_desc == "identity_scalar doc_out"

        info = proxy.get_command_config()
        assert isinstance(info, CommandInfoList)
        assert len(info) == 5
        assert info[3].cmd_name == "identity_scalar"
        assert info[3].disp_level == DispLevel.OPERATOR
        assert info[3].in_type == DevLong64
        assert info[3].in_type_desc == "identity_scalar doc_in"
        assert info[3].out_type == DevLong64
        assert info[3].out_type_desc == "identity_scalar doc_out"

        assert info[4].cmd_name == "identity_spectrum"
        assert info[4].disp_level == DispLevel.EXPERT
        assert info[4].in_type == DevVarDoubleArray
        assert info[4].in_type_desc == "identity_spectrum doc_in"
        assert info[4].out_type == DevVarDoubleArray
        assert info[4].out_type_desc == "identity_spectrum doc_out"


def test_docstring_parsing():

    pytest.importorskip("docstring_parser")

    class TestDevice(Device):

        @command
        def rest_docstring(self, val_in: int) -> int:
            """
            :param val_in: The input integer value.
            :type val_in: int
            :return: An integer result after processing.
            :rtype: int
            """
            return val_in

        @command
        def rest_docstring_only_in(self, val_in: int):
            """
            :param val_in: The input integer value.
            :type val_in: int
            """
            pass

        @command
        def rest_docstring_only_out(self) -> int:
            """
            :return: An integer result after processing.
            :rtype: int
            """
            pass

        @command
        def rest_docstring_args_list_only_out(self, *args) -> int:
            """
            :return: An integer result after processing.
            :rtype: int
            """
            pass

        @command
        def google_docstring(self, val_in: int) -> int:
            """
            Args:
                val_in (int): The input integer value.

            Returns:
                int: An integer result after processing.
            """
            return val_in

        @command
        def numpy_docstring(self, val_in: int) -> int:
            """
            Parameters
            ----------
            val_in : int
                The input integer value.

            Returns
            -------
            int
                An integer result after processing.
            """
            return val_in

        @command
        def epydoc_docstring(self, val_in: int) -> int:
            """
            @param val_in: The input integer value.
            @type val_in: int
            @return: An integer result after processing.
            @rtype: int
            """
            return val_in

        @command(
            doc_in="val_in (int): The input integer value.",
            doc_out="returns (int): An integer result after processing.",
        )
        def kwargs_have_priority_over_docstring(self, val_in: int) -> int:
            """
            :param val_in: THIS DESCRIPTION WON'T BE USED
            :type val_in: int
            :return: THIS DESCRIPTION WON'T BE USED
            :rtype: int
            """
            return val_in

        @command
        def no_docstring(self, val_in: int) -> int:
            return val_in

        @command
        def no_docstring_args_list_in(self, *args) -> int:
            return 123

        @command
        def no_docstring_only_in(self, val_in: int): ...

        @command
        def no_docstring_only_out(self) -> int:
            return 123

        @command
        def no_docstring_no_in_no_out(self): ...

    with DeviceTestContext(TestDevice) as proxy:

        expected_int_in_desc = "val_in (int): The input integer value."
        expected_int_out_desc = "returns (int): An integer result after processing."
        expected_void_in_desc = "No input parameter (DevVoid)"
        expected_void_out_desc = "No output parameter (DevVoid)"

        for cmd_name in [
            "rest_docstring",
            "google_docstring",
            "numpy_docstring",
            "epydoc_docstring",
            "kwargs_have_priority_over_docstring",
        ]:
            cfg = proxy.get_command_config(cmd_name)
            assert cfg.in_type_desc == expected_int_in_desc
            assert cfg.out_type_desc == expected_int_out_desc

        cfg = proxy.get_command_config("rest_docstring_only_in")
        assert cfg.in_type_desc == expected_int_in_desc
        assert cfg.out_type_desc == expected_void_out_desc

        cfg = proxy.get_command_config("rest_docstring_only_out")
        assert cfg.in_type_desc == expected_void_in_desc
        assert cfg.out_type_desc == expected_int_out_desc

        expected_int_in_desc = (
            ":param val_in: (not documented)\n" ":type val_in: DevLong64"
        )
        expected_int_out_desc = ":return: (not documented)\n:rtype: DevLong64"

        cfg = proxy.get_command_config("no_docstring")
        assert cfg.in_type_desc == expected_int_in_desc
        assert cfg.out_type_desc == expected_int_out_desc

        cfg = proxy.get_command_config("no_docstring_args_list_in")
        assert cfg.in_type_desc == expected_void_in_desc
        assert cfg.out_type_desc == expected_int_out_desc

        cfg = proxy.get_command_config("no_docstring_only_in")
        assert cfg.in_type_desc == expected_int_in_desc
        assert cfg.out_type_desc == expected_void_out_desc

        cfg = proxy.get_command_config("no_docstring_only_out")
        assert cfg.in_type_desc == expected_void_in_desc
        assert cfg.out_type_desc == expected_int_out_desc

        cfg = proxy.get_command_config("no_docstring_no_in_no_out")
        assert cfg.in_type_desc == expected_void_in_desc
        assert cfg.out_type_desc == expected_void_out_desc


def test_fill_cmd_polling_buffer(command_typed_values):
    dtype, values, expected = command_typed_values

    start_time = time.time()

    class TestDevice(Device):

        @command(dtype_out=dtype)
        def cmd(self):
            return values[0]

        @command
        def fill_history(self):
            data = []
            for i, val in enumerate(values):
                data.append(TimedCmdData(val, time_stamp=start_time + i))

            data.append(
                TimedCmdData(
                    error=RuntimeError("Test"),
                    time_stamp=start_time + len(data),
                )
            )
            self.fill_cmd_polling_buffer("cmd", data)

    with DeviceTestContext(TestDevice) as proxy:
        # we do not want, that tango core fills up history automatically
        proxy.poll_command("cmd", 0)
        proxy.fill_history()
        history = proxy.command_history("cmd", 10)
        assert len(history) == len(values) + 1
        for ind, cmd in enumerate(history[:-1]):
            assert_close(cmd.extract(), expected(values[ind]))
            assert_close(start_time + ind, cmd.get_date().totime())

        assert history[-1].has_failed()


def test_decorated_void_command():

    def mutate_command(method):
        @wraps(method)
        def wrapper(self):
            _ = method(self, "added arg")
            return

        return wrapper

    class TestDevice(Device):

        @command(dtype_in=None, dtype_out=None)
        @mutate_command
        def none_as_void_cmd(self, arg: str) -> str:
            assert arg == "added arg"
            return "some ret"

        @command(dtype_in=DevVoid, dtype_out=DevVoid)
        @mutate_command
        def devvoid_cmd(self, arg: str) -> str:
            assert arg == "added arg"
            return "some ret"

        @command(dtype_in="DevVoid", dtype_out="DevVoid")
        @mutate_command
        def str_devvoid_cmd(self, arg: str) -> str:
            assert arg == "added arg"
            return "some ret"

    with DeviceTestContext(TestDevice) as proxy:
        proxy.none_as_void_cmd()
        proxy.devvoid_cmd()
        proxy.str_devvoid_cmd()



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""
This is an internal PyTango module.
"""

__all__ = ("pytango_pprint_init",)

__docformat__ = "restructuredtext"

import json
import textwrap

from tango import (
    StdStringVector,
    StdLongVector,
    CommandInfoList,
    AttributeInfoList,
    AttributeInfoListEx,
    DeviceDataHistoryList,
    GroupReplyList,
    GroupAttrReplyList,
    GroupCmdReplyList,
    DbData,
    DbDevInfos,
    DbDevExportInfos,
    DbDevImportInfos,
    DbHistoryList,
    LockerInfo,
    DevCommandInfo,
    AttributeDimension,
    CommandInfo,
    DeviceInfo,
    DeviceAttributeConfig,
    AttributeInfo,
    AttributeAlarmInfo,
    ChangeEventInfo,
    PeriodicEventInfo,
    ArchiveEventInfo,
    AttributeEventInfo,
    AttributeInfoEx,
    DeviceAttribute,
    DeviceAttributeHistory,
    DeviceData,
    DeviceDataHistory,
    DbDatum,
    DbDevInfo,
    DbDevImportInfo,
    DbDevExportInfo,
    DbServerInfo,
    GroupReply,
    GroupAttrReply,
    GroupCmdReply,
    DevError,
    EventData,
    AttrConfEventData,
    DataReadyEventData,
    TimeVal,
    DevFailed,
    CmdArgType,
    MultiAttrProp,
    ClientAddr,
)

from tango._tango import (
    AttributeConfig,
    AttributeConfig_2,
    AttributeConfig_3,
    AttributeConfig_5,
    AttributeAlarm,
    EventProperties,
    ChangeEventProp,
    PeriodicEventProp,
    ArchiveEventProp,
)
import collections.abc


_INDENT_LEVEL = 4
_INDENT = " " * _INDENT_LEVEL
_STRUCT_TYPES = (
    LockerInfo,
    DevCommandInfo,
    AttributeDimension,
    CommandInfo,
    DeviceInfo,
    DeviceAttributeConfig,
    AttributeInfo,
    AttributeAlarmInfo,
    ChangeEventInfo,
    PeriodicEventInfo,
    ArchiveEventInfo,
    AttributeEventInfo,
    AttributeInfoEx,
    DeviceAttribute,
    DeviceAttributeHistory,
    DeviceData,
    DeviceDataHistory,
    DbDatum,
    DbDevInfo,
    DbDevImportInfo,
    DbDevExportInfo,
    DbServerInfo,
    DevError,
    EventData,
    AttrConfEventData,
    DataReadyEventData,
    AttributeConfig,
    AttributeConfig_2,
    AttributeConfig_3,
    AttributeConfig_5,
    ChangeEventProp,
    PeriodicEventProp,
    ArchiveEventProp,
    AttributeAlarm,
    EventProperties,
    MultiAttrProp,
)
_SEQUENCE_TYPES = (
    StdStringVector,
    StdLongVector,
    CommandInfoList,
    AttributeInfoList,
    AttributeInfoListEx,
    DeviceDataHistoryList,
    GroupReplyList,
    GroupAttrReplyList,
    GroupCmdReplyList,
    DbData,
    DbDevInfos,
    DbDevExportInfos,
    DbDevImportInfos,
    DbHistoryList,
)


def __inc_param(obj, name):
    ret = not name.startswith("_")
    ret &= name not in ("except_flags",)
    ret &= not isinstance(getattr(obj, name), collections.abc.Callable)
    return ret


def __nested_json_like_repr(value) -> str:
    if isinstance(value, str):
        return f'"{value}"'
    elif type(value) in _STRUCT_TYPES:
        return str(value)
    elif isinstance(value, dict):
        try:
            return json.dumps(value, indent=_INDENT_LEVEL, sort_keys=True)
        except TypeError:
            return repr(value)
    else:
        return repr(value)


def __single_param(obj, param_name, f=repr, fmt="%s = %s"):
    param_value = getattr(obj, param_name)
    if param_name == "data_type":
        param_value = CmdArgType.values.get(param_value, param_value)
    return fmt % (param_name, f(param_value))


def __struct_params_s(obj, separator=", ", f=repr, fmt="%s = %s"):
    """method wrapper for printing all elements of a struct"""
    s = separator.join(
        [__single_param(obj, n, f, fmt) for n in dir(obj) if __inc_param(obj, n)]
    )
    return s


def __struct_params_repr(obj):
    """method wrapper for representing all elements of a struct"""
    return __struct_params_s(obj)


def __struct_params_str(obj, fmt, f=repr):
    """method wrapper for printing all elements of a struct."""
    return __struct_params_s(obj, "\n", f=f, fmt=fmt)


def __repr__Struct(self):
    """repr method for struct"""
    return f"{self.__class__.__name__}({__struct_params_repr(self)})"


def __str__Struct(self):
    """str method for struct"""
    fmt = "%s = %s"
    details = __struct_params_str(self, fmt, __nested_json_like_repr)
    details = __indented(details, strip_outer=False)
    result = f"{self.__class__.__name__}[\n{details}\n]"
    return result


def __registerSeqStr():
    """helper function to make internal sequences printable"""
    _SeqStr = lambda self: (self and f"[{', '.join(map(repr, self))}]") or "[]"
    _SeqRepr = lambda self: (self and f"[{', '.join(map(repr, self))}]") or "[]"

    for seq in _SEQUENCE_TYPES:
        seq.__str__ = _SeqStr
        seq.__repr__ = _SeqRepr


def __str__DevFailed(self):
    if isinstance(self.args, collections.abc.Sequence):
        seq_str = __str_error_stack_helper(self.args)
        return f"DevFailed[\n{seq_str}\n]"
    return f"DevFailed[{self.args}]"


def __str_error_stack_helper(errors):
    err_str = ",\n".join(str(err).strip() for err in errors)
    err_str = __indented(err_str, strip_outer=False)
    return err_str


def __repr__DevFailed(self):
    return f"DevFailed(args = {repr(self.args)})"


def __str__DevError(self):
    details = (
        f"desc = {__indented(self.desc)}\n"
        f"origin = {__indented(self.origin)}\n"
        f"reason = {self.reason}\n"
        f"severity = {self.severity}\n"
    )
    details = __indented(details, strip_outer=False)
    s = f"DevError[\n{details}\n]"
    return s


def __indented(text, strip_outer=True):
    indented = textwrap.indent(text.strip(), _INDENT)
    if strip_outer:
        return indented.strip()
    else:
        return indented


def __registerStructStr():
    """helper method to register str and repr methods for structures"""

    for struct in _STRUCT_TYPES:
        struct.__str__ = __str__Struct
        struct.__repr__ = __repr__Struct

    # special case for structs that already have a str representation
    TimeVal.__repr__ = __repr__Struct
    GroupReply.__repr__ = __repr__Struct
    GroupAttrReply.__repr__ = __repr__Struct
    GroupCmdReply.__repr__ = __repr__Struct
    ClientAddr.__repr__ = __repr__Struct

    # special case for DevFailed: we want a better pretty print
    # also, because it is an Exception it has the message attribute which
    # generates a Deprecation warning in python 2.6
    DevFailed.__str__ = __str__DevFailed
    DevFailed.__repr__ = __repr__DevFailed

    DevError.__str__ = __str__DevError


def pytango_pprint_init():
    __registerSeqStr()
    __registerStructStr()



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
import asyncio
from tango.asyncio import DeviceProxy


async def asyncio_example():
    dev = await DeviceProxy("sys/tg_test/1")
    print(dev.get_green_mode())

    print(await dev.state())

    # in case of high-level API read has to be awaited
    print(await dev.long_scalar)
    print(await dev["long_scalar"])
    print(await getattr(dev, "long_scalar"))

    # while write executed sync
    dev.long_scalar = 1

    # for low-level API both read_attribute and write_attribute have to be awaited
    print(await dev.read_attribute("long_scalar"))
    await dev.write_attribute("long_scalar", 1)


if __name__ == "__main__":
    asyncio.run(asyncio_example())



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later


import os

# Concurrent imports
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor

# Tango imports
from tango.green import AbstractExecutor, get_ident
from tango.utils import _get_current_otel_context, _get_non_tango_source_location

__all__ = (
    "FuturesExecutor",
    "get_global_executor",
    "set_global_executor",
    "_switch_global_executor_to_thread",
)

# Global executor

_MAIN_EXECUTOR = None
_THREAD_EXECUTORS = {}


def _switch_global_executor_to_thread():
    """
    internal PyTango function, use only if you sure, what you are doing!
    Used for correct behavior of TestDeviceContext
    checks, that global executor belongs to the caller thread, and,
    if not - creates a new one and saves it as a new global
    """
    global _MAIN_EXECUTOR
    if _MAIN_EXECUTOR is not None and not _MAIN_EXECUTOR.in_executor_context():
        # we save current executor in the known subthread executors to be used later
        _THREAD_EXECUTORS[_MAIN_EXECUTOR.get_ident()] = _MAIN_EXECUTOR
        _MAIN_EXECUTOR = FuturesExecutor()


def get_global_executor():
    global _MAIN_EXECUTOR
    if _MAIN_EXECUTOR is None:
        _MAIN_EXECUTOR = FuturesExecutor()

    # the following patch is used for correct behavior of TestDeviceContext,
    # which has two different executors for main and device threads
    if not _MAIN_EXECUTOR.in_executor_context():
        ident = get_ident(), os.getpid()
        if ident in _THREAD_EXECUTORS:
            return _THREAD_EXECUTORS[ident]

    return _MAIN_EXECUTOR


def set_global_executor(executor):
    global _MAIN_EXECUTOR
    _MAIN_EXECUTOR = executor


# Futures executor


class FuturesExecutor(AbstractExecutor):
    """Futures tango executor"""

    asynchronous = True
    default_wait = True

    def __init__(self, process=False, max_workers=20):
        super().__init__()
        cls = ProcessPoolExecutor if process else ThreadPoolExecutor
        self.subexecutor = cls(max_workers=max_workers)

    def delegate(self, fn, *args, **kwargs):
        """Return the given operation as a concurrent future."""
        if hasattr(fn, "__trace_kwargs__"):
            kwargs["trace_location"] = _get_non_tango_source_location()
            kwargs["trace_context"] = _get_current_otel_context()
        return self.subexecutor.submit(fn, *args, **kwargs)

    def access(self, accessor, timeout=None):
        """Return a result from a single callable."""
        return accessor.result(timeout=timeout)

    def submit(self, fn, *args, **kwargs):
        """Submit an operation"""
        return fn(*args, **kwargs)

    def execute(self, fn, *args, **kwargs):
        """Execute an operation and return the result."""
        return fn(*args, **kwargs)



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""
This is an internal PyTango module.
"""


__all__ = ("AttrData",)

__docformat__ = "restructuredtext"

import inspect

from tango._tango import Except, CmdArgType, AttrDataFormat, AttrWriteType
from tango._tango import DispLevel, UserDefaultAttrProp, UserDefaultFwdAttrProp
from tango._tango import Attr, SpectrumAttr, ImageAttr, FwdAttr
from tango.utils import is_non_str_seq, is_pure_str


class AttrData:
    """A helper class that contains the same information one of the items in
    DeviceClass.attr_list but in object form"""

    def __init__(self, name, class_name, attr_info=None):
        self.class_name = class_name
        self.attr_name = name
        self.attr_type = CmdArgType.DevVoid
        self.attr_format = AttrDataFormat.SCALAR
        self.attr_write = AttrWriteType.READ
        self.dim_x = 1
        self.dim_y = 0
        self.display_level = DispLevel.OPERATOR
        self.polling_period = -1
        self.memorized = False
        self.hw_memorized = False
        self.alarm_event_implemented = False
        self.alarm_event_detect = True
        self.archive_event_implemented = False
        self.archive_event_detect = True
        self.change_event_implemented = False
        self.change_event_detect = True
        self.data_ready_event_implemented = False
        if name is None:
            self.read_method_name = None
            self.write_method_name = None
            self.is_allowed_name = None
        else:
            self.read_method_name = "read_" + name
            self.write_method_name = "write_" + name
            self.is_allowed_name = "is_" + name + "_allowed"
        self.attr_class = None
        self.attr_args = None
        self.att_prop = None
        self.forward = False
        self.has_dtype_kword = False
        self.has_size_kword = False
        if attr_info is not None:
            self.from_attr_info(attr_info)

    @classmethod
    def from_dict(cls, attr_dict):
        attr_dict = dict(attr_dict)
        name = attr_dict.pop("name", None)
        class_name = attr_dict.pop("class_name", None)
        self = cls(name, class_name)
        self.build_from_dict(attr_dict)
        return self

    def build_from_dict(self, attr_dict):
        self.forward = attr_dict.pop("forwarded", False)
        if not self.forward:
            self.has_dtype_kword = "dtype" in attr_dict
            self.attr_type = attr_dict.pop("dtype", CmdArgType.DevDouble)
            self.attr_format = attr_dict.pop("dformat", AttrDataFormat.SCALAR)
            self.has_size_kword = "max_dim_x" in attr_dict or "max_dim_y" in attr_dict
            self.dim_x = attr_dict.pop("max_dim_x", 1)
            self.dim_y = attr_dict.pop("max_dim_y", 0)
            self.display_level = attr_dict.pop("display_level", DispLevel.OPERATOR)
            self.polling_period = attr_dict.pop("polling_period", -1)
            self.memorized = attr_dict.pop("memorized", False)
            self.hw_memorized = attr_dict.pop("hw_memorized", False)

            is_access_explicit = "access" in attr_dict
            if is_access_explicit:
                self.attr_write = attr_dict.pop("access")
            else:
                # access is defined by which methods were defined
                r_explicit = "fread" in attr_dict or "fget" in attr_dict
                w_explicit = "fwrite" in attr_dict or "fset" in attr_dict
                if r_explicit and w_explicit:
                    self.attr_write = AttrWriteType.READ_WRITE
                elif r_explicit:
                    self.attr_write = AttrWriteType.READ
                elif w_explicit:
                    self.attr_write = AttrWriteType.WRITE
                else:
                    self.attr_write = AttrWriteType.READ

            fread = attr_dict.pop("fget", attr_dict.pop("fread", None))
            if fread is not None:
                if is_pure_str(fread):
                    self.read_method_name = fread
                elif inspect.isroutine(fread):
                    self.read_method_name = fread.__name__
            fwrite = attr_dict.pop("fset", attr_dict.pop("fwrite", None))
            if fwrite is not None:
                if is_pure_str(fwrite):
                    self.write_method_name = fwrite
                elif inspect.isroutine(fwrite):
                    self.write_method_name = fwrite.__name__
            fisallowed = attr_dict.pop("fisallowed", None)
            if fisallowed is not None:
                if is_pure_str(fisallowed):
                    self.is_allowed_name = fisallowed
                elif inspect.isroutine(fisallowed):
                    self.is_allowed_name = fisallowed.__name__
            self.attr_class = attr_dict.pop("klass", None)
        else:
            self.attr_class = FwdAttr
            self.attr_args = [self.attr_name]

        self.alarm_event_implemented = attr_dict.pop("alarm_event_implemented", False)
        self.alarm_event_detect = attr_dict.pop("alarm_event_detect", True)
        self.archive_event_implemented = attr_dict.pop(
            "archive_event_implemented", False
        )
        self.archive_event_detect = attr_dict.pop("archive_event_detect", True)
        self.change_event_implemented = attr_dict.pop("change_event_implemented", False)
        self.change_event_detect = attr_dict.pop("change_event_detect", True)
        self.data_ready_event_implemented = attr_dict.pop(
            "data_ready_event_implemented", False
        )

        if self.forward:
            self.att_prop = self.__create_user_default_fwdattr_prop(attr_dict)
        elif len(attr_dict):
            self.att_prop = self.__create_user_default_attr_prop(attr_dict)
        return self

    def _set_name(self, name):
        old_name = self.attr_name
        self.attr_name = name
        if self.attr_args is None:
            self.attr_args = [name]
        else:
            self.attr_args[0] = name
        if old_name is None:
            if self.read_method_name is None:
                self.read_method_name = "read_" + name
            if self.write_method_name is None:
                self.write_method_name = "write_" + name
            if self.is_allowed_name is None:
                self.is_allowed_name = "is_" + name + "_allowed"

    def __throw_exception(self, msg, meth="create_attribute()"):
        Except.throw_exception("PyDs_WrongAttributeDefinition", msg, meth)

    def __create_user_default_fwdattr_prop(self, extra_info):
        """for internal usage only"""
        p = UserDefaultFwdAttrProp()
        if "label" in extra_info:
            p.set_label(extra_info["label"])
        return p

    def __create_user_default_attr_prop(self, extra_info):
        """for internal usage only"""
        p = UserDefaultAttrProp()

        doc = extra_info.pop("doc", None)
        if doc is not None:
            extra_info["description"] = doc

        for k, v in extra_info.items():
            k_lower = k.lower()
            method_name = f"set_{k_lower.replace(' ', '_')}"
            if hasattr(p, method_name):
                method = getattr(p, method_name)
                if method_name == "set_enum_labels":
                    method(v)
                else:
                    method(str(v))
            elif k == "delta_time":
                p.set_delta_t(str(v))
            elif k_lower not in ("display level", "polling period", "memorized"):
                msg = (
                    f"Wrong definition of attribute. "
                    f"The object extra information '{k}' "
                    f"is not recognized!"
                )
                Except.throw_exception(
                    "PyDs_WrongAttributeDefinition",
                    msg,
                    "create_user_default_attr_prop()",
                )
        return p

    def from_attr_info(self, attr_info):
        self.has_dtype_kword = True
        self.has_size_kword = True

        name = self.class_name
        attr_name = self.attr_name
        throw_ex = self.__throw_exception
        # check for well defined attribute info

        # check parameter
        if not is_non_str_seq(attr_info):
            throw_ex(
                f"Wrong data type for value for describing attribute {attr_name} in "
                f"class {name}\nMust be a sequence with 1 or 2 elements"
            )

        if len(attr_info) < 1 or len(attr_info) > 2:
            throw_ex(
                f"Wrong number of argument for describing attribute {attr_name} in "
                f"class {name}\nMust be a sequence with 1 or 2 elements"
            )

        extra_info = {}
        if len(attr_info) == 2:
            # attr_info[1] must be a dictionary
            # extra_info = attr_info[1], with all the keys lowercase
            for k, v in attr_info[1].items():
                extra_info[k.lower()] = v

        attr_info = attr_info[0]

        attr_info_len = len(attr_info)
        # check parameter
        if not is_non_str_seq(attr_info) or attr_info_len < 3 or attr_info_len > 5:
            throw_ex(
                f"Wrong data type for describing mandatory information for "
                f"attribute {attr_name} in class {name}\nMust be a sequence with 3, 4 "
                f"or 5 elements"
            )

        # get data type
        try:
            self.attr_type = CmdArgType(attr_info[0])
        except Exception:
            throw_ex(
                f"Wrong data type in attribute argument for attribute {attr_name} "
                f"in class {name}\nAttribute data type (first element in first "
                f"sequence) must be a tango.CmdArgType"
            )

        # get format
        try:
            self.attr_format = AttrDataFormat(attr_info[1])
        except Exception:
            throw_ex(
                f"Wrong data format in attribute argument for attribute {attr_name} "
                f"in class {name}\nAttribute data format (second element in "
                f"first sequence) must be a tango.AttrDataFormat"
            )

        if self.attr_format == AttrDataFormat.SCALAR:
            if attr_info_len != 3:
                throw_ex(
                    f"Wrong data type in attribute argument for attribute "
                    f"{attr_name} in class {name}\nSequence describing mandatory "
                    f"attribute parameters for scalar attribute must have "
                    f"3 elements"
                )
        elif self.attr_format == AttrDataFormat.SPECTRUM:
            if attr_info_len != 4:
                throw_ex(
                    f"Wrong data type in attribute argument for attribute "
                    f"{attr_name} in class {name}\nSequence describing mandatory "
                    f"attribute parameters for spectrum attribute must "
                    f"have 4 elements"
                )
            try:
                self.dim_x = int(attr_info[3])
            except Exception:
                throw_ex(
                    f"Wrong data type in attribute argument for attribute "
                    f"{attr_name} in class {name}\n4th element in sequence describing "
                    f"mandatory dim_x attribute parameter for spectrum "
                    f"attribute must be an integer"
                )
        elif self.attr_format == AttrDataFormat.IMAGE:
            if attr_info_len != 5:
                throw_ex(
                    f"Wrong data type in attribute argument for attribute "
                    f"{attr_name} in class {name}\nSequence describing mandatory "
                    f"attribute parameters for image attribute must have "
                    f"5 elements"
                )
            try:
                self.dim_x = int(attr_info[3])
            except Exception:
                throw_ex(
                    f"Wrong data type in attribute argument for attribute "
                    f"{attr_name} in class {name}\n4th element in sequence describing "
                    f"mandatory dim_x attribute parameter for image "
                    f"attribute must be an integer"
                )
            try:
                self.dim_y = int(attr_info[4])
            except Exception:
                throw_ex(
                    f"Wrong data type in attribute argument for attribute "
                    f"{attr_name} in class {name}\n5th element in sequence desribing "
                    f"mandatory dim_y attribute parameter for image "
                    f"attribute must be an integer"
                )
        else:
            assert False, "Unsupported AttrDataFormat {attr_format}."

        # get write type
        try:
            self.attr_write = AttrWriteType(attr_info[2])
        except Exception:
            throw_ex(
                f"Wrong data write type in attribute argument for "
                f"attribute {attr_name} in class {name}\nAttribute write type (third "
                f"element in first sequence) must be a "
                f"tango.AttrWriteType"
            )
        try:
            self.display_level = DispLevel(
                extra_info.get("display level", DispLevel.OPERATOR)
            )
        except Exception:
            throw_ex(
                f"Wrong display level in attribute information for "
                f"attribute {attr_name} in class {name}\nAttribute information for "
                f"display level is not a tango.DispLevel"
            )
        try:
            self.polling_period = int(extra_info.get("polling period", -1))
        except Exception:
            throw_ex(
                f"Wrong polling period in attribute information for "
                f"attribute {attr_name} in class {name}\nAttribute information for "
                f"polling period is not an integer"
            )

        memorized = extra_info.get("memorized", "false").lower()

        if memorized not in ("true", "true_without_hard_applied", "false"):
            throw_ex(
                f"Wrong memorized value for attribute {attr_name} in class {name}."
                f'Allowed valued are the strings "true", "false" and '
                f'"true_without_hard_applied" (case insensitive)'
            )

        if memorized == "true":
            self.memorized = True
            self.hw_memorized = True
        elif memorized == "true_without_hard_applied":
            self.memorized = True
        else:
            self.memorized = False

        self.alarm_event_implemented = extra_info.get("alarm_event_implemented", False)
        self.alarm_event_detect = extra_info.get("alarm_event_detect", True)
        self.archive_event_implemented = extra_info.get(
            "archive_event_implemented", False
        )
        self.archive_event_detect = extra_info.get("archive_event_detect", True)
        self.change_event_implemented = extra_info.get(
            "change_event_implemented", False
        )
        self.change_event_detect = extra_info.get("change_event_detect", True)
        self.data_ready_event_implemented = extra_info.get(
            "data_ready_event_implemented", False
        )

        if self.attr_type == CmdArgType.DevEnum:
            if "enum_labels" not in extra_info:
                throw_ex(
                    f"Missing 'enum_labels' key in attr_list definition "
                    f"for enum attribute {attr_name} in class {name}"
                )
            self.enum_labels = extra_info["enum_labels"]

        self.attr_class = extra_info.get("klass", None)

        att_prop = None
        if extra_info:
            att_prop = self.__create_user_default_attr_prop(extra_info)
        self.att_prop = att_prop

    def set_enum_labels_to_attr_prop(self, enum_labels):
        if self.att_prop:
            self.att_prop.set_enum_labels(enum_labels)
        else:
            self.att_prop = self.__create_user_default_attr_prop(
                {"enum_labels": enum_labels}
            )

    def to_attr(self):
        if self.attr_args is None:
            attr_args = [self.attr_name, self.attr_type, self.attr_write]
            if not self.attr_format == AttrDataFormat.SCALAR:
                attr_args.append(self.dim_x)
                if not self.attr_format == AttrDataFormat.SPECTRUM:
                    attr_args.append(self.dim_y)
        else:
            attr_args = self.attr_args

        if self.attr_class is None:
            klass = self.DftAttrClassMap[self.attr_format]
        else:
            klass = self.attr_class

        attr = klass(*attr_args)
        if self.att_prop is not None:
            attr.set_default_properties(self.att_prop)
        attr.set_disp_level(self.display_level)
        if self.memorized:
            attr.set_memorized()
            attr.set_memorized_init(self.hw_memorized)
        if self.polling_period > 0:
            attr.set_polling_period(self.polling_period)
        return attr

    DftAttrClassMap = {
        AttrDataFormat.SCALAR: Attr,
        AttrDataFormat.SPECTRUM: SpectrumAttr,
        AttrDataFormat.IMAGE: ImageAttr,
    }



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
import asyncio
import inspect
import multiprocessing
import os
import sys
import textwrap
import threading
import time

import numpy as np

try:
    import numpy.typing as npt
except ImportError:
    npt = None

import pytest

from collections.abc import Callable

import tango.asyncio
import tango.constants
from tango import (
    AttrData,
    Attr,
    AttrDataFormat,
    AttReqType,
    AttrWriteType,
    CmdArgType,
    DevFailed,
    DevState,
    DeviceClass,
    DeviceProxy,
    EventType,
    GreenMode,
    LatestDeviceImpl,
    EnsureOmniThread,
    LockerLanguage,
)
from tango.server import __to_callback, BaseDevice, Device
from tango.pyutil import parse_args
from tango.server import command, attribute, class_property, device_property
from tango.test_utils import (
    DeviceTestContext,
    MultiDeviceTestContext,
    GoodEnum,
    BadEnumNonZero,
    BadEnumSkipValues,
    BadEnumDuplicates,
    DEVICE_SERVER_ARGUMENTS,
    wait_for_proxy,
)
from tango.utils import (
    EnumTypeError,
    get_enum_labels,
    get_latest_device_class,
    is_pure_str,
    get_tango_type_format,
    parse_type_hint,
)


# Constants
TIMEOUT = 10.0

# Test implementation classes

WRONG_HINTS = (  # hint_caller, type_hint, error_reason
    ("property", tuple[tuple[int]], "Property does not support IMAGE type"),
    (
        "property",
        tuple[tuple[int, float], float],
        "Property does not support IMAGE type",
    ),
    ("property", tuple[int, float], "PyTango does not support mixed types"),
    ("attribute", tuple[int, float], "PyTango does not support mixed types"),
    (
        "attribute",
        tuple[tuple[int, float], float],
        "PyTango does not support mixed types",
    ),
    (
        "attribute",
        tuple[tuple[int, int], list[int, int]],
        "PyTango does not support mixed types",
    ),
    ("attribute", Callable[[int], None], "Cannot translate"),
)


@pytest.mark.parametrize("hint_caller, type_hint, error_reason", WRONG_HINTS)
def test_uncorrect_typing_hints(hint_caller, type_hint, error_reason):
    with pytest.raises(RuntimeError, match=error_reason):
        dtype, dformat, max_x, max_y = parse_type_hint(type_hint, caller=hint_caller)
        get_tango_type_format(dtype, dformat, hint_caller)


@pytest.fixture(params=[GoodEnum])
def good_enum(request):
    return request.param


@pytest.fixture(params=[BadEnumNonZero, BadEnumSkipValues, BadEnumDuplicates])
def bad_enum(request):
    return request.param


# test utilities for servers


def test_get_enum_labels_success(good_enum):
    expected_labels = ["START", "MIDDLE", "END"]
    assert get_enum_labels(good_enum) == expected_labels


def test_get_enum_labels_fail(bad_enum):
    with pytest.raises(EnumTypeError):
        get_enum_labels(bad_enum)


def test_device_classes_use_latest_implementation():
    assert issubclass(LatestDeviceImpl, get_latest_device_class())
    assert issubclass(BaseDevice, LatestDeviceImpl)
    assert issubclass(Device, BaseDevice)


# Test state/status


def test_empty_device(server_green_mode):
    class TestDevice(Device):
        green_mode = server_green_mode

    with DeviceTestContext(TestDevice) as proxy:
        assert proxy.state() == DevState.UNKNOWN
        assert proxy.status() == "The device is in UNKNOWN state."


@pytest.mark.parametrize("description_source", ["doc", "description"])
def test_set_desc_status_state_at_init(description_source):
    class TestDevice(Device):
        if description_source == "doc":
            __doc__ = "Test name"
        else:
            # device_class_description has priority
            __doc__ = "Test name 2"
            DEVICE_CLASS_DESCRIPTION = "Test name"
        DEVICE_CLASS_INITIAL_STATUS = "Test status"
        DEVICE_CLASS_INITIAL_STATE = DevState.ON

    class ChildDevice(TestDevice):
        pass

    class SecondChildDevice(TestDevice):
        DEVICE_CLASS_DESCRIPTION = "Test name 2"
        DEVICE_CLASS_INITIAL_STATUS = "Test status 2"
        DEVICE_CLASS_INITIAL_STATE = DevState.OFF

    devices_info = (
        {"class": TestDevice, "devices": [{"name": "test/dev/main"}]},
        {"class": ChildDevice, "devices": [{"name": "test/dev/child1"}]},
        {"class": SecondChildDevice, "devices": [{"name": "test/dev/child2"}]},
    )

    with MultiDeviceTestContext(devices_info) as context:
        for proxy in [
            context.get_device("test/dev/main"),
            context.get_device("test/dev/child1"),
        ]:
            assert proxy.state() == DevState.ON
            assert proxy.status() == "Test status"
            if (
                description_source == "description"
            ):  # note, that docsrting is not inherited!
                assert proxy.description() == "Test name"

        proxy = context.get_device("test/dev/child2")
        assert proxy.state() == DevState.OFF
        assert proxy.status() == "Test status 2"
        assert proxy.description() == "Test name 2"


@pytest.mark.parametrize("force_user_status", [False, True])
def test_set_state_status(state, server_green_mode, force_user_status):
    if force_user_status:
        status = "\n".join(
            (
                "This is a multiline status",
                "with special characters such as",
                "Café à la crème",
            )
        )
    else:
        status = f"The device is in {state} state."

    if server_green_mode == GreenMode.Asyncio:

        class TestDevice(Device):
            green_mode = server_green_mode

            async def init_device(self):
                self.set_state(state)
                if force_user_status:
                    self.set_status(status)

    else:

        class TestDevice(Device):
            green_mode = server_green_mode

            def init_device(self):
                self.set_state(state)
                if force_user_status:
                    self.set_status(status)

    with DeviceTestContext(TestDevice) as proxy:
        assert proxy.state() == state
        assert proxy.status() == status


def test_user_dev_state_status(server_green_mode):
    state = DevState.MOVING
    status = "Device is MOVING"

    if server_green_mode == GreenMode.Asyncio:

        class TestDevice(Device):
            green_mode = server_green_mode

            async def dev_state(self):
                return state

            async def dev_status(self):
                return status

    else:

        class TestDevice(Device):
            green_mode = server_green_mode

            def dev_state(self):
                return state

            def dev_status(self):
                return status

    with DeviceTestContext(TestDevice) as proxy:
        assert proxy.state() == state
        assert proxy.status() == status


def test_attr_quality_checked_with_state(server_green_mode):
    if server_green_mode == GreenMode.Asyncio:

        class BaseTestDevice(Device):
            @command(dtype_out=bool)
            async def check_sub_function_was_called(self):
                return (
                    self.read_attr_hardware_was_called
                    and self.always_executed_hook_was_called
                )

    else:

        class BaseTestDevice(Device):
            @command(dtype_out=bool)
            def check_sub_function_was_called(self):
                return (
                    self.read_attr_hardware_was_called
                    and self.always_executed_hook_was_called
                )

    class TestDevice(BaseTestDevice):
        green_mode = server_green_mode

        read_attr_hardware_was_called = False
        always_executed_hook_was_called = False

        sync_code = textwrap.dedent(
            """
            def init_device(self):
                Device.init_device(self)
                self.set_state(DevState.ON)

            def read_attr_hardware(self, attr_list):
                self.read_attr_hardware_was_called = True
                return Device.read_attr_hardware(self, attr_list)

            def always_executed_hook(self):
                self.always_executed_hook_was_called = True

            @attribute(max_alarm=0)
            def test_attribute(self):
                return 42
                """
        )

        if server_green_mode == GreenMode.Asyncio:
            exec(
                sync_code.replace("def", "async def").replace("Device", "await Device")
            )
        else:
            exec(sync_code)

    with DeviceTestContext(TestDevice) as proxy:
        assert proxy.state() == DevState.ALARM
        assert proxy.check_sub_function_was_called()


def test_device_get_attr_config(server_green_mode):
    class TestDevice(Device):
        # green mode matters to check deadlocks in async modes
        green_mode = server_green_mode

        sync_code = textwrap.dedent(
            """
        @attribute(dtype=bool)
        def attr_config_ok(self):
            # testing that call to get_attribute_config for all types of
            # input arguments gives same result and doesn't raise an exception
            ac1 = self.get_attribute_config("attr_config_ok")
            ac2 = self.get_attribute_config(["attr_config_ok"])
            return repr(ac1) == repr(ac2)
        """
        )

        if server_green_mode == GreenMode.Asyncio:
            exec(sync_code.replace("def", "async def"))
        else:
            exec(sync_code)

    with DeviceTestContext(TestDevice) as proxy:
        assert proxy.attr_config_ok


def test_device_set_attr_config(server_green_mode):
    class TestDevice(Device):
        # green mode matters to check deadlocks in async modes
        green_mode = server_green_mode

        sync_code = textwrap.dedent(
            """
        @attribute(dtype=int)
        def attr(self):
            attr_config = self.get_attribute_config("attr")
            attr_config[0].min_value = "-7"
            attr_config[0].min_alarm = "-6"

            attr_config[0].max_alarm = "6"
            attr_config[0].max_value = "7"

            self.set_attribute_config(attr_config)
            assert repr(attr_config) == repr(self.get_attribute_config("attr"))

            with pytest.raises(AttributeError, match="object has no attribute 'lala'"):
                attr_config[0].lala = "7"

            attr_config = self.get_attribute_config_3("attr")
            attr_config[0].min_value = "-5"
            attr_config[0].att_alarm.min_alarm = "-4"
            attr_config[0].att_alarm.min_warning = "-3"

            attr_config[0].att_alarm.max_warning = "3"
            attr_config[0].att_alarm.max_alarm = "4"
            attr_config[0].max_value = "5"

            self.set_attribute_config_3(attr_config)
            assert repr(attr_config) == repr(self.get_attribute_config_3("attr"))

            with pytest.raises(AttributeError, match="object has no attribute 'lala'"):
                attr_config[0].lala = "7"

            attr = self.get_device_attr().get_attr_by_name("attr")

            val = -2
            for f in ["min_alarm", "min_warning", "max_warning", "max_alarm"]:
                getattr(attr, f"set_{f}")(val)
                assert val == getattr(attr, f"get_{f}")()
                val += 1

            return 1
            """
        )

        if server_green_mode == GreenMode.Asyncio:
            exec(sync_code.replace("def", "async def"))
        else:
            exec(sync_code)

    with DeviceTestContext(TestDevice) as proxy:
        assert proxy.attr == 1


def test_default_units():
    # testing that, by default tango.constants.UnitNotSpec is set
    # when no unit is specified. For bool, int, float and str dtypes
    class TestDevice(Device):

        @attribute(dtype=bool)
        def attr_bool_ok(self):
            return True

        @attribute(dtype=int)
        def attr_int_ok(self):
            return 1

        @attribute(dtype=float)
        def attr_float_ok(self):
            return 1.0

        @attribute(dtype=str)
        def attr_str_ok(self):
            return "True"

    def assert_attr_bool_ok(dev_proxy):
        config = dev_proxy.get_attribute_config("attr_bool_ok")
        assert config.unit == tango.constants.UnitNotSpec

    def assert_attr_int_ok(dev_proxy):
        config = dev_proxy.get_attribute_config("attr_int_ok")
        assert config.unit == tango.constants.UnitNotSpec

    def assert_attr_float_ok(dev_proxy):
        config = dev_proxy.get_attribute_config("attr_float_ok")
        assert config.unit == tango.constants.UnitNotSpec

    def assert_attr_str_ok(dev_proxy):
        config = dev_proxy.get_attribute_config("attr_str_ok")
        assert config.unit == tango.constants.UnitNotSpec

    with DeviceTestContext(TestDevice) as proxy:
        assert_attr_bool_ok(proxy)
        assert_attr_int_ok(proxy)
        assert_attr_float_ok(proxy)
        assert_attr_str_ok(proxy)


def test_custom_units():
    class TestDevice(Device):

        @attribute(dtype=bool, unit="mA")
        def custom_unit_ok(self):
            return True

    def assert_custom_unit_ok(dev_proxy):
        config = dev_proxy.get_attribute_config("custom_unit_ok")
        assert config.unit == "mA"

    with DeviceTestContext(TestDevice) as proxy:
        assert_custom_unit_ok(proxy)


# Test inheritance


def test_inheritance_overrides_a_property():

    class A(Device):

        dev_prop1 = device_property(dtype=str, default_value="hello_dev1")
        dev_prop2 = device_property(dtype=str, default_value="hello_dev2")
        class_prop1 = class_property(dtype=str, default_value="hello_class1")
        class_prop2 = class_property(dtype=str, default_value="hello_class2")

        @command(dtype_out=str)
        def get_dev_prop1(self):
            return self.dev_prop1

        @command(dtype_out=str)
        def get_dev_prop2(self):
            return self.dev_prop2

        @command(dtype_out=str)
        def get_class_prop1(self):
            return self.class_prop1

        @command(dtype_out=str)
        def get_class_prop2(self):
            return self.class_prop2

    class B(A):
        dev_prop2 = device_property(dtype=str, default_value="goodbye_dev2")
        class_prop2 = class_property(dtype=str, default_value="goodbye_class2")

    devices_info = (
        {"class": A, "devices": [{"name": "test/dev/a"}]},
        {"class": B, "devices": [{"name": "test/dev/b"}]},
    )

    with MultiDeviceTestContext(devices_info) as context:
        proxy_a = context.get_device("test/dev/a")
        proxy_b = context.get_device("test/dev/b")

        assert proxy_a.get_dev_prop1() == "hello_dev1"
        assert proxy_a.get_dev_prop2() == "hello_dev2"
        assert proxy_a.get_class_prop1() == "hello_class1"
        assert proxy_a.get_class_prop2() == "hello_class2"

        assert proxy_b.get_dev_prop1() == "hello_dev1"
        assert proxy_b.get_dev_prop2() == "goodbye_dev2"
        assert proxy_b.get_class_prop1() == "hello_class1"
        assert proxy_b.get_class_prop2() == "goodbye_class2"


def test_inheritance_override_dev_status():
    class A(Device):

        def dev_status(self):
            return ")`'-.,_"

    class B(A):
        def dev_status(self):
            return 3 * A.dev_status(self)

    with DeviceTestContext(B) as proxy:
        assert proxy.status() == ")`'-.,_)`'-.,_)`'-.,_"


def test_inheritance_init_device():

    class A(Device):
        initialised_count_a = 0

        def init_device(self):
            super().init_device()
            self.initialised_count_a += 1

        @command(dtype_out=int)
        def get_is_initialised_a(self):
            return self.initialised_count_a

    class B(A):
        initialised_count_b = 0

        def init_device(self):
            super().init_device()
            self.initialised_count_b += 1

        @command(dtype_out=int)
        def get_is_initialised_b(self):
            return self.initialised_count_b

    with DeviceTestContext(B) as proxy:
        assert proxy.get_is_initialised_a() == 1
        assert proxy.get_is_initialised_b() == 1


def test_inheritance_with_decorated_attributes():
    is_allowed = True

    class A(Device):

        @attribute(access=AttrWriteType.READ_WRITE)
        def decorated_a(self):
            return self.decorated_a_value

        @decorated_a.setter
        def decorated_a(self, value):
            self.decorated_a_value = value

        @decorated_a.is_allowed
        def decorated_a(self, req_type):
            assert req_type in (AttReqType.READ_REQ, AttReqType.WRITE_REQ)
            return is_allowed

    class B(A):
        @attribute(access=AttrWriteType.READ_WRITE)
        def decorated_b(self):
            return self.decorated_b_value

        @decorated_b.setter
        def decorated_b(self, value):
            self.decorated_b_value = value

        @decorated_b.is_allowed
        def decorated_b(self, req_type):
            assert req_type in (AttReqType.READ_REQ, AttReqType.WRITE_REQ)
            return is_allowed

    with DeviceTestContext(B) as proxy:
        is_allowed = True

        proxy.decorated_a = 1.23
        assert proxy.decorated_a == 1.23
        proxy.decorated_b = 4.5
        assert proxy.decorated_b == 4.5

        is_allowed = False
        with pytest.raises(DevFailed):
            proxy.decorated_a = 1.0
        with pytest.raises(DevFailed):
            _ = proxy.decorated_a
        with pytest.raises(DevFailed):
            proxy.decorated_b = 1.0
        with pytest.raises(DevFailed):
            _ = proxy.decorated_b


def test_inheritance_with_undecorated_attributes():
    is_allowed = True

    class A(Device):

        attr_a = attribute(access=AttrWriteType.READ_WRITE)

        def _check_is_allowed(self):
            return is_allowed

        def read_attr_a(self):
            return self.attr_a_value

        def write_attr_a(self, value):
            self.attr_a_value = value

        def is_attr_a_allowed(self, req_type):
            assert req_type in (AttReqType.READ_REQ, AttReqType.WRITE_REQ)
            return self._check_is_allowed()

    class B(A):
        attr_b = attribute(access=AttrWriteType.READ_WRITE)

        def read_attr_b(self):
            return self.attr_b_value

        def write_attr_b(self, value):
            self.attr_b_value = value

        def is_attr_b_allowed(self, req_type):
            assert req_type in (AttReqType.READ_REQ, AttReqType.WRITE_REQ)
            return self._check_is_allowed()

    with DeviceTestContext(B) as proxy:
        is_allowed = True

        proxy.attr_a = 2.5
        assert proxy.attr_a == 2.5
        proxy.attr_b = 5.75
        assert proxy.attr_b == 5.75

        is_allowed = False
        with pytest.raises(DevFailed):
            proxy.attr_a = 1.0
        with pytest.raises(DevFailed):
            _ = proxy.attr_a
        with pytest.raises(DevFailed):
            proxy.attr_b = 1.0
        with pytest.raises(DevFailed):
            _ = proxy.attr_b


def test_inheritance_with_undecorated_attribute_and_bound_methods():

    class A(Device):

        is_allowed = True

        attr_a = attribute(
            access=AttrWriteType.READ_WRITE,
            fget="get_attr_a",
            fset="set_attr_a",
            fisallowed="isallowed_attr_a",
        )

        def get_attr_a(self):
            return self.attr_value

        def set_attr_a(self, value):
            self.attr_value = value

        def isallowed_attr_a(self, req_type):
            assert req_type in (AttReqType.READ_REQ, AttReqType.WRITE_REQ)
            return self.is_allowed

        @command(dtype_in=bool)
        def make_allowed(self, yesno):
            self.is_allowed = yesno

    class B(A):
        attr_b = attribute(
            access=AttrWriteType.READ_WRITE,
            fget="get_attr_b",
            fset="set_attr_b",
            fisallowed="isallowed_attr_b",
        )

        def get_attr_b(self):
            return self.attr_value

        def set_attr_b(self, value):
            self.attr_value = value

        def isallowed_attr_b(self, req_type):
            assert req_type in (AttReqType.READ_REQ, AttReqType.WRITE_REQ)
            return self.is_allowed

    with DeviceTestContext(B) as proxy:
        proxy.attr_a = 3.75
        assert proxy.attr_a == 3.75
        proxy.attr_b = 6.0
        assert proxy.attr_b == 6.0

        proxy.make_allowed(False)
        with pytest.raises(DevFailed):
            proxy.attr_a = 1.0
        with pytest.raises(DevFailed):
            _ = proxy.attr_a
        with pytest.raises(DevFailed):
            proxy.attr_b = 1.0
        with pytest.raises(DevFailed):
            _ = proxy.attr_b


def test_inheritance_with_undecorated_attributes_and_unbound_functions():
    is_allowed = True
    values = {"a": 0.0, "b": 0.0}

    def read_attr_a(device):
        assert isinstance(device, B)
        return values["a"]

    def write_attr_a(device, value):
        assert isinstance(device, B)
        values["a"] = value

    def is_attr_a_allowed(device, req_type):
        assert isinstance(device, B)
        assert req_type in (AttReqType.READ_REQ, AttReqType.WRITE_REQ)
        return is_allowed

    class A(Device):

        attr_a = attribute(
            access=AttrWriteType.READ_WRITE,
            fget=read_attr_a,
            fset=write_attr_a,
            fisallowed=is_attr_a_allowed,
        )

    def read_attr_b(device):
        assert isinstance(device, B)
        return values["b"]

    def write_attr_b(device, value):
        assert isinstance(device, B)
        values["b"] = value

    def is_attr_b_allowed(device, req_type):
        assert isinstance(device, B)
        assert req_type in (AttReqType.READ_REQ, AttReqType.WRITE_REQ)
        return is_allowed

    class B(A):
        attr_b = attribute(
            access=AttrWriteType.READ_WRITE,
            fget=read_attr_b,
            fset=write_attr_b,
            fisallowed=is_attr_b_allowed,
        )

    with DeviceTestContext(B) as proxy:
        is_allowed = True

        proxy.attr_a = 2.5
        assert proxy.attr_a == 2.5
        proxy.attr_b = 5.75
        assert proxy.attr_b == 5.75

        is_allowed = False
        with pytest.raises(DevFailed):
            proxy.attr_a = 1.0
        with pytest.raises(DevFailed):
            _ = proxy.attr_a
        with pytest.raises(DevFailed):
            proxy.attr_b = 1.0
        with pytest.raises(DevFailed):
            _ = proxy.attr_b


def test_inheritance_command_is_allowed_by_naming_convention():

    class A(Device):

        @command(dtype_out=str)
        def cmd(self):
            return "ok"

        def is_cmd_allowed(self):
            return is_allowed

    class B(A):
        pass

    with DeviceTestContext(B) as proxy:
        is_allowed = True
        assert proxy.cmd() == "ok"
        is_allowed = False
        with pytest.raises(DevFailed):
            proxy.cmd()


def test_inheritance_command_is_allowed_by_kwarg_method():

    class A(Device):

        @command(dtype_out=str, fisallowed="fisallowed_kwarg_method")
        def cmd(self):
            return "ok 1"

        def fisallowed_kwarg_method(self):
            return is_allowed

    class B(A):
        pass

    with DeviceTestContext(B) as proxy:
        is_allowed = True
        assert proxy.cmd() == "ok 1"
        is_allowed = False
        with pytest.raises(DevFailed):
            proxy.cmd()


def test_inheritance_command_is_allowed_by_kwarg_unbound_function():
    is_allowed = True

    def fisallowed_function(self):
        return is_allowed

    class A(Device):
        @command(dtype_out=str, fisallowed=fisallowed_function)
        def cmd(self):
            return "ok"

    class B(A):
        pass

    with DeviceTestContext(B) as proxy:
        is_allowed = True
        assert proxy.cmd() == "ok"
        is_allowed = False
        with pytest.raises(DevFailed):
            proxy.cmd()


# Test Exception propagation
def test_exception_propagation(server_green_mode):
    if server_green_mode == GreenMode.Asyncio:

        class TestDevice(Device):
            green_mode = server_green_mode

            @attribute
            async def attr(self):
                1 / 0  # pylint: disable=pointless-statement

            @command
            async def cmd(self):
                1 / 0  # pylint: disable=pointless-statement

    else:

        class TestDevice(Device):
            green_mode = server_green_mode

            @attribute
            def attr(self):
                1 / 0  # pylint: disable=pointless-statement

            @command
            def cmd(self):
                1 / 0  # pylint: disable=pointless-statement

    with DeviceTestContext(TestDevice) as proxy:
        with pytest.raises(DevFailed) as record:
            proxy.attr  # pylint: disable=pointless-statement
        assert "ZeroDivisionError" in record.value.args[0].desc

        with pytest.raises(DevFailed) as record:
            proxy.cmd()
        assert "ZeroDivisionError" in record.value.args[0].desc


def _avoid_double_colon_node_ids(val):
    """Return node IDs without a double colon.

    IDs with "::" can't be used to launch a test from the command line, as pytest
    considers this sequence as a module/test name separator.  Add something extra
    to keep them usable for single test command line execution (e.g., under Windows CI).
    """
    if is_pure_str(val) and "::" in val:
        return str(val).replace("::", ":_:")


@pytest.fixture(params=["linux", "win"])
def os_system(request):
    original_platform = sys.platform
    sys.platform = request.param
    yield
    sys.platform = original_platform


@pytest.mark.parametrize(
    "applicable_os, test_input, expected_output",
    DEVICE_SERVER_ARGUMENTS,
    ids=_avoid_double_colon_node_ids,
)
def test_arguments(applicable_os, test_input, expected_output, os_system):
    try:
        assert set(parse_args(test_input.split())) == set(expected_output)
    except SystemExit:
        assert sys.platform not in applicable_os


# Test Server init hook


def test_server_init_hook_called(server_green_mode):
    if server_green_mode == GreenMode.Asyncio:

        class TestDevice(Device):
            green_mode = server_green_mode
            server_init_hook_called = False

            async def server_init_hook(self):
                await asyncio.sleep(0.01)
                TestDevice.server_init_hook_called = True

    else:

        class TestDevice(Device):
            green_mode = server_green_mode
            server_init_hook_called = False

            def server_init_hook(self):
                TestDevice.server_init_hook_called = True

    with DeviceTestContext(TestDevice):
        assert TestDevice.server_init_hook_called


def test_server_init_hook_change_state():

    class TestDevice(Device):
        def server_init_hook(self):
            self.set_state(DevState.ON)

    with DeviceTestContext(TestDevice) as proxy:
        assert proxy.state() == DevState.ON


def test_asyncio_server_init_hook_change_state():
    class TestDevice(Device):
        green_mode = GreenMode.Asyncio

        async def server_init_hook(self):
            await asyncio.sleep(0.01)
            self.set_state(DevState.ON)

    with DeviceTestContext(TestDevice) as proxy:
        assert proxy.state() == DevState.ON


def test_server_init_hook_called_after_init():
    class TestDevice(Device):
        def init_device(self):
            self.set_state(DevState.INIT)

        def server_init_hook(self):
            self.set_state(DevState.ON)

    with DeviceTestContext(TestDevice) as proxy:
        assert proxy.state() == DevState.ON


def test_async_server_init_hook_called_after_init():
    class TestDevice(Device):
        green_mode = GreenMode.Asyncio

        async def init_device(self):
            await asyncio.sleep(0.01)
            self.set_state(DevState.INIT)

        async def server_init_hook(self):
            await asyncio.sleep(0.01)
            self.set_state(DevState.ON)

    with DeviceTestContext(TestDevice) as proxy:
        assert proxy.state() == DevState.ON


def test_server_init_hook_exception():
    class TestDevice(Device):

        def server_init_hook(self):
            self.set_state(DevState.ON)
            raise RuntimeError("Force exception for test")

    with DeviceTestContext(TestDevice) as proxy:
        assert proxy.state() == DevState.FAULT


def test_asyncio_server_init_hook_exception():
    class TestDevice(Device):
        green_mode = GreenMode.Asyncio

        async def server_init_hook(self):
            await asyncio.sleep(0.01)
            raise RuntimeError("Force exception for test")

    with DeviceTestContext(TestDevice) as proxy:
        assert proxy.state() == DevState.FAULT


def test_server_init_hook_with_low_level_api_called():
    class ClassicAPISimpleDeviceImpl(LatestDeviceImpl):
        has_been_called = False

        def server_init_hook(self):
            self.set_state(DevState.ON)
            ClassicAPISimpleDeviceImpl.has_been_called = True

    class ClassicAPISimpleDeviceClass(DeviceClass):
        pass

    with DeviceTestContext(ClassicAPISimpleDeviceImpl, ClassicAPISimpleDeviceClass):
        assert ClassicAPISimpleDeviceImpl.has_been_called


def test_server_init_hook_with_low_level_api_change_state():
    class ClassicAPISimpleDeviceImpl(LatestDeviceImpl):

        def server_init_hook(self):
            self.set_state(DevState.ON)

    class ClassicAPISimpleDeviceClass(DeviceClass):
        pass

    with DeviceTestContext(
        ClassicAPISimpleDeviceImpl, ClassicAPISimpleDeviceClass
    ) as proxy:
        assert proxy.state() == DevState.ON


def test_server_init_hook_with_low_level_api_called_after_init():
    class ClassicAPISimpleDeviceImpl(LatestDeviceImpl):
        def init_device(self):
            self.set_state(DevState.INIT)

        def server_init_hook(self):
            self.set_state(DevState.ON)

    class ClassicAPISimpleDeviceClass(DeviceClass):
        pass

    with DeviceTestContext(
        ClassicAPISimpleDeviceImpl, ClassicAPISimpleDeviceClass
    ) as proxy:
        assert proxy.state() == DevState.ON


def test_server_init_hook_with_low_level_api_exception():
    class ClassicAPISimpleDeviceImpl(LatestDeviceImpl):

        def server_init_hook(self):
            self.set_state(DevState.ON)
            raise RuntimeError("Force exception for test")

    class ClassicAPISimpleDeviceClass(DeviceClass):
        pass

    with DeviceTestContext(
        ClassicAPISimpleDeviceImpl, ClassicAPISimpleDeviceClass
    ) as proxy:
        assert proxy.state() == DevState.FAULT


def test_server_init_multiple_devices():
    event_list = []

    class DeviceOne(Device):
        def server_init_hook(self):
            event_list.append("DeviceOne")

    class DeviceTwo(Device):
        def server_init_hook(self):
            event_list.append("DeviceTwo")

    devices_info = (
        {"class": DeviceOne, "devices": [{"name": "test/device1/1"}]},
        {
            "class": DeviceTwo,
            "devices": [{"name": "test/device2/1"}, {"name": "test/device3/1"}],
        },
    )

    with MultiDeviceTestContext(devices_info):
        assert len(event_list) == 3
        assert "DeviceOne" in event_list
        assert "DeviceTwo" in event_list


def test_server_init_hook_subscribe_event_multiple_devices():
    pytest.xfail("This test is unreliable - to be fixed soon")

    event_queue = multiprocessing.Queue()

    class DeviceOne(Device):
        @attribute(dtype=int)
        def some_attribute(self):
            return 42

        def init_device(self):
            super().init_device()
            self.set_change_event("some_attribute", implemented=True, detect=False)

        @command
        def push_event_cmd(self):
            self.push_change_event("some_attribute", 43)

    class DeviceTwo(Device):
        def event_handler(self, data):
            event_queue.put(data.attr_value.value)

        def server_init_hook(self):
            self.dev1_proxy = DeviceProxy("test/device1/1")
            self.dev1_proxy.subscribe_event(
                "some_attribute", EventType.CHANGE_EVENT, self.event_handler
            )

    devices_info = (
        {"class": DeviceOne, "devices": [{"name": "test/device1/1"}]},
        {
            "class": DeviceTwo,
            "devices": [{"name": "test/device2/1"}, {"name": "test/device3/1"}],
        },
    )

    with MultiDeviceTestContext(devices_info) as context:
        proxy = context.get_device("test/device1/1")

        # synchronous event
        assert 42 == event_queue.get(timeout=TIMEOUT)
        assert 42 == event_queue.get(timeout=TIMEOUT)
        assert event_queue.empty()

        # asynchronous event pushed from user code
        proxy.push_event_cmd()
        assert 43 == event_queue.get(timeout=TIMEOUT)
        assert 43 == event_queue.get(timeout=TIMEOUT)
        assert event_queue.empty()


def test_deprecation_warning_for_sync_attr_com_methods_in_asyncio_device():
    class TestDevice(Device):
        green_mode = GreenMode.Asyncio
        attr_value = 1

        # static attributes and commands

        @attribute(access=AttrWriteType.READ_WRITE)
        async def attr_all_methods_async(self) -> int:
            return self.attr_value

        @attr_all_methods_async.write
        async def attr_all_methods_async(self, value):
            self.attr_value = value

        @attr_all_methods_async.is_allowed
        async def attr_all_methods_async(self, req_type):
            return True

        @attribute(access=AttrWriteType.READ_WRITE)
        def attr_sync_read_write(self) -> int:
            return self.attr_value

        @attr_sync_read_write.write
        def set_attr_sync_read_write(self, value):
            self.attr_value = value

        @attribute
        async def attr_sync_is_allowed(self) -> int:
            return self.attr_value

        @attr_sync_is_allowed.is_allowed
        def is_attr_sync_is_allowed(self, req_type):
            return True

        @command(dtype_out=int)
        async def cmd_all_methods_async(self, val_in: int) -> int:
            return val_in

        async def is_cmd_all_methods_async_allowed(self):
            return True

        @command(dtype_out=int)
        def cmd_sync_func(self, val_in: int) -> int:
            return val_in

        @command(dtype_out=int)
        async def cmd_sync_is_allowed(self, val_in: int) -> int:
            return val_in

        def is_cmd_sync_is_allowed_allowed(self):
            return True

        # dynamic attributes and commands

        @command
        async def add_dynamic_cmd_attr(self):
            attr = attribute(
                name="dyn_attr_all_methods_async",
                access=AttrWriteType.READ_WRITE,
                fget=self.dyn_attr_all_methods_async,
                fset=self.dyn_set_attr_all_methods_async,
                fisallowed=self.is_dyn_attr_all_methods_async_allowed,
            )
            self.add_attribute(attr)

            attr = attribute(
                name="dyn_attr_sync_read_write",
                access=AttrWriteType.READ_WRITE,
                fget=self.dyn_attr_sync_read_write,
                fset=self.dyn_set_attr_sync_read_write,
            )
            self.add_attribute(attr)

            attr = attribute(
                name="dyn_attr_sync_is_allowed",
                access=AttrWriteType.READ,
                fget=self.dyn_attr_sync_is_allowed,
                fisallowed=self.is_dyn_attr_sync_is_allowed,
            )
            self.add_attribute(attr)

            cmd = command(
                f=self.dyn_cmd_all_methods_async,
                fisallowed=self.is_dyn_cmd_all_methods_async_allowed,
            )
            self.add_command(cmd)

            cmd = command(f=self.dyn_cmd_sync_func)
            self.add_command(cmd)

            cmd = command(
                f=self.dyn_cmd_sync_is_allowed,
                fisallowed=self.is_dyn_cmd_sync_is_allowed_allowed,
            )
            self.add_command(cmd)

        async def dyn_attr_all_methods_async(self, attr) -> int:
            return self.attr_value

        async def dyn_set_attr_all_methods_async(self, attr):
            self.attr_value = attr.get_write_value()

        async def is_dyn_attr_all_methods_async_allowed(self, req_type):
            return True

        def dyn_attr_sync_read_write(self, attr) -> int:
            return self.attr_value

        def dyn_set_attr_sync_read_write(self, attr):
            self.attr_value = attr.get_write_value()

        async def dyn_attr_sync_is_allowed(self, attr) -> int:
            return self.attr_value

        def is_dyn_attr_sync_is_allowed(self, req_type):
            return True

        async def dyn_cmd_all_methods_async(self, val_in: int) -> int:
            return val_in

        async def is_dyn_cmd_all_methods_async_allowed(self):
            return True

        def dyn_cmd_sync_func(self, val_in: int) -> int:
            return val_in

        async def dyn_cmd_sync_is_allowed(self, val_in: int) -> int:
            return val_in

        def is_dyn_cmd_sync_is_allowed_allowed(self):
            return True

    with DeviceTestContext(TestDevice) as proxy:
        proxy.add_dynamic_cmd_attr()

        proxy.attr_all_methods_async = 123
        assert proxy.attr_all_methods_async == 123

        proxy.dyn_attr_all_methods_async = 456
        assert proxy.dyn_attr_all_methods_async == 456

        with pytest.warns(DeprecationWarning):
            proxy.attr_sync_read_write = 123

        with pytest.warns(DeprecationWarning):
            assert proxy.attr_sync_read_write == 123

        with pytest.warns(DeprecationWarning):
            assert proxy.attr_sync_is_allowed == 123

        with pytest.warns(DeprecationWarning):
            proxy.dyn_attr_sync_read_write = 456

        with pytest.warns(DeprecationWarning):
            assert proxy.dyn_attr_sync_read_write == 456

        with pytest.warns(DeprecationWarning):
            assert proxy.dyn_attr_sync_is_allowed == 456

        assert proxy.cmd_all_methods_async(123) == 123

        with pytest.warns(DeprecationWarning):
            assert proxy.cmd_sync_func(123) == 123

        with pytest.warns(DeprecationWarning):
            assert proxy.cmd_sync_is_allowed(123) == 123

        assert proxy.dyn_cmd_all_methods_async(123) == 123

        with pytest.warns(DeprecationWarning):
            assert proxy.dyn_cmd_sync_func(123) == 123

        with pytest.warns(DeprecationWarning):
            assert proxy.dyn_cmd_sync_is_allowed(123) == 123


@pytest.mark.parametrize(
    "method",
    [
        "init_device",
        "delete_device",
        "dev_state",
        "dev_status",
        "read_attr_hardware",
        "always_executed_hook",
    ],
)
def test_deprecation_warning_for_standard_methods_in_asyncio_device(method):
    class TestDevice(Device):
        green_mode = GreenMode.Asyncio

        @attribute
        async def attr(self) -> int:
            return 1

        async_code = textwrap.dedent(
            """
            async def init_device(self):
                pass

            async def delete_device(self):
                pass

            async def dev_state(self):
                return DevState.ON

            async def dev_status(self):
                return "All good"

            async def read_attr_hardware(self, attr_list):
                pass

            async def always_executed_hook(self):
                 pass
             """
        )

        exec(async_code.replace(f"async def {method}", f"def {method}"))

    with pytest.warns(DeprecationWarning, match=method):
        with DeviceTestContext(TestDevice) as proxy:
            _ = proxy.state()
            _ = proxy.status()
            _ = proxy.attr


@pytest.mark.skip(
    reason="This test fails because the first attempt to solve this problem caused a regression and the MR was reverted"
)
def test_no_sync_attribute_locks(server_green_mode):
    """
    Without AttributeMonitor locks, reading attributes while
    simultaneously pushing change events would crash the device
    in NO_SYNC modes: Asyncio and Gevent.
    """

    class BaseTestDevice(Device):
        def __init__(self, *args):
            super().__init__(*args)
            self._last_data = 0.0
            self._publisher = threading.Thread(
                target=self._publisher_thread, name="publisher"
            )
            self._publisher.daemon = True
            self._running = False
            self.set_change_event("H22", implemented=True, detect=False)

        def _publisher_thread(self):
            with EnsureOmniThread():
                while self._running:
                    self._last_data = np.random.rand()
                    super().push_change_event("H22", self._last_data)

    if server_green_mode == GreenMode.Asyncio:

        class TestDevice(BaseTestDevice):
            green_mode = server_green_mode

            @command
            async def Start(self):
                self._running = True
                self._publisher.start()

            @command
            async def Stop(self):
                self._running = False

            @attribute(dtype=float)
            async def H22(self):
                return self._last_data

    else:

        class TestDevice(BaseTestDevice):
            green_mode = server_green_mode

            @command
            def Start(self):
                self._running = True
                self._publisher.start()

            @command
            def Stop(self):
                self._running = False

            @attribute(dtype=float)
            def H22(self):
                return self._last_data

    with DeviceTestContext(TestDevice) as proxy:
        proxy.Start()
        # This loop should be enough to crash the device
        # with previous unpatched code in 99% of the cases
        for _ in range(15):
            proxy.H22
        proxy.Stop()


def test_read_slow_and_fast_attributes_with_asyncio():
    class MyDevice(Device):
        green_mode = GreenMode.Asyncio

        @attribute(dtype=str)
        async def slow(self):
            await asyncio.sleep(1)
            return "slow"

        @attribute(dtype=str)
        async def fast(self):
            return "fast"

    context = DeviceTestContext(MyDevice)
    context.start()
    access = context.get_device_access()
    read_order = []

    def read_slow_attribute():
        proxy = DeviceProxy(access)
        read_order.append(proxy.slow)

    def read_fast_attribute():
        proxy = DeviceProxy(access)
        read_order.append(proxy.fast)

    slow_thread = threading.Thread(target=read_slow_attribute)
    fast_thread = threading.Thread(target=read_fast_attribute)
    slow_thread.start()
    time.sleep(0.5)
    fast_thread.start()

    slow_thread.join()
    fast_thread.join()
    context.stop()

    assert read_order == ["fast", "slow"]


def test_get_version_info_classic_api():
    version_info = dict()

    class ClassicAPIDeviceImpl(LatestDeviceImpl):
        def __init__(self, cl, name):
            super().__init__(cl, name)
            ClassicAPIDeviceImpl.init_device(self)

        def init_device(self):
            version_info.update(self.get_version_info())

    class ClassicAPIClass(DeviceClass):
        pass

    with DeviceTestContext(ClassicAPIDeviceImpl, ClassicAPIClass) as proxy:
        assert "PyTango" in version_info
        assert "NumPy" in version_info
        assert proxy.info().version_info == version_info


def test_get_version_info_high_level_api():
    version_info = dict()

    class TestDevice(Device):
        def init_device(self):
            version_info.update(self.get_version_info())

    with DeviceTestContext(TestDevice) as proxy:
        assert "PyTango" in version_info
        assert "NumPy" in version_info
        assert proxy.info().version_info == version_info


def test_add_version_info_classic_api():
    class ClassicAPIDeviceImpl(LatestDeviceImpl):
        def __init__(self, cl, name):
            super().__init__(cl, name)
            ClassicAPIDeviceImpl.init_device(self)

        def init_device(self):
            self.add_version_info("device_version", "1.0.0")

    class ClassicAPIClass(DeviceClass):
        pass

    with DeviceTestContext(ClassicAPIDeviceImpl, ClassicAPIClass) as proxy:
        assert proxy.info().version_info["device_version"] == "1.0.0"


def test_add_version_info_high_level_api():
    class TestDevice(Device):
        def init_device(self):
            self.add_version_info("device_version", "1.0.0")

    with DeviceTestContext(TestDevice) as proxy:
        assert proxy.info().version_info["device_version"] == "1.0.0"


@pytest.mark.extra_src_test
def test_restart_server_command_cpp_and_py(mixed_tango_test_server):
    process, proxy_when_ready = mixed_tango_test_server

    proxy = proxy_when_ready()
    assert proxy.state() == DevState.ON

    proxy.command_inout("RestartServer")

    # after restart the proxy is unavailable for a short time, so we wait again
    proxy = proxy_when_ready()
    time.sleep(0.1)  # give TangoTest some extra time to start

    assert proxy.state() == DevState.ON

    # terminate early so we can verify that there is a clean exit
    process.terminate()
    process.join(timeout=3.0)  # Allow TangoTest time to stop DataGenerator

    assert not process.is_alive()
    assert process.exitcode == 0


def test_restart_simple_server_does_not_crash(server_green_mode):

    class TestDevice(Device):
        pass

    devices_info = ({"class": TestDevice, "devices": [{"name": "device/test/1"}]},)
    with MultiDeviceTestContext(devices_info) as context:
        proxy = context.get_device(context.server_name)
        assert proxy.state() == DevState.ON
        proxy.command_inout("RestartServer")
        time.sleep(0.1)
        new_proxy = wait_for_proxy(context.server_name)
        assert new_proxy.state() == DevState.ON
        assert proxy.state() == DevState.ON


def test_attr_data_default_fwd_properties():

    attr_name = "some_attr"
    class_name = "some_class"
    label_value = "abcd"

    d = {
        "name": attr_name,
        "class_name": class_name,
        "forwarded": True,
        "label": label_value,
    }
    AttrData.from_dict(d)
    # can't query "label" as that is in a private section in cppTango's UserDefaultFwdAttrProp
    pass


def test_attr_data_default_properties():

    attr_name = "some_attr"
    class_name = "some_class"

    d = {
        "name": attr_name,
        "class_name": class_name,
        "forwarded": False,
        "delta_time": 1234,
    }

    attr_data = AttrData.from_dict(d)
    assert attr_data.att_prop.delta_t == "1234"


def test_attr_data_default_properties_throws_on_unknown():

    attr_name = "some_attr"
    class_name = "some_class"

    d = {
        "name": attr_name,
        "class_name": class_name,
        "forwarded": False,
        "memorized": True,
        "I DONT EXIST": None,
    }

    with pytest.raises(DevFailed, match="Wrong definition of attribute"):
        AttrData.from_dict(d)


def test_attr_data_enum_labels():

    attr_name = "some_attr"
    class_name = "some_class"
    label_value_1 = "abcd"
    label_value_2 = "efgh"

    d = {
        "name": attr_name,
        "class_name": class_name,
    }
    attr_data = AttrData.from_dict(d)

    attr_data.set_enum_labels_to_attr_prop([label_value_1])
    assert attr_data.att_prop is not None
    assert attr_data.att_prop.enum_labels == label_value_1

    # set different enum values
    attr_data.set_enum_labels_to_attr_prop([label_value_2])
    assert attr_data.att_prop.enum_labels == label_value_2


def test_attr_data_to_attr():

    attr_name = "some_attr"
    class_name = "some class"
    poll_period = 123

    d = {
        "name": attr_name,
        "class_name": class_name,
        "klass": Attr,
        "memorized": True,
        "fread": "read_method",
        "fwrite": "write_method",
        "polling_period": poll_period,
    }

    attr_data = AttrData.from_dict(d)

    assert attr_data.attr_args is None
    assert attr_data.attr_class is not None

    attr = attr_data.to_attr()

    assert attr.get_memorized()
    assert attr.get_polling_period() == poll_period

    # attr_args != None case
    attr_data.attr_args = [
        attr_data.attr_name,
        attr_data.attr_type,
        attr_data.attr_write,
    ]

    attr = attr_data.to_attr()

    assert attr.get_memorized()
    assert attr.get_polling_period() == poll_period


@pytest.mark.parametrize(
    "attr_info,exception_message",
    [
        ("some string", "Wrong data type for value for describing attribute"),
        ([], "Wrong number of argument for describing attribute"),
        ([1, 2, 3], "Wrong number of argument for describing attribute"),
        ([[1]], "Wrong data type for describing mandatory information"),
        ([[1, 2, 3, 4, 5, 6]], "Wrong data type for describing mandatory information"),
        (
            [["abcd", "abcd", "abcd"]],
            "Wrong data type in attribute argument for attribute",
        ),
        (
            [[CmdArgType.DevDouble, "abcd", "abcd"]],
            "Wrong data format in attribute argument for attribute",
        ),
        (
            [[CmdArgType.DevDouble, AttrDataFormat.SCALAR, "abcd", "abcd"]],
            "Sequence describing mandatory attribute parameters for scalar attribute must have 3 elements",
        ),
        (
            [[CmdArgType.DevDouble, AttrDataFormat.SPECTRUM, "abcd"]],
            "Sequence describing mandatory attribute parameters for spectrum attribute must have 4 elements",
        ),
        (
            [[CmdArgType.DevDouble, AttrDataFormat.SPECTRUM, "abcd", "abcd"]],
            "mandatory dim_x attribute parameter for spectrum attribute must be an integer",
        ),
        (
            [[CmdArgType.DevDouble, AttrDataFormat.IMAGE, "abcd", "abcd"]],
            "Sequence describing mandatory attribute parameters for image attribute must have 5 elements",
        ),
        (
            [[CmdArgType.DevDouble, AttrDataFormat.IMAGE, "abcd", "abcd", "abcd"]],
            "mandatory dim_x attribute parameter for image attribute must be an integer",
        ),
        (
            [[CmdArgType.DevDouble, AttrDataFormat.IMAGE, "abcd", 1, "abcd"]],
            "mandatory dim_y attribute parameter for image attribute must be an integer",
        ),
        (
            [[CmdArgType.DevDouble, AttrDataFormat.SCALAR, "abcd"]],
            "Wrong data write type in attribute argument",
        ),
        (
            [
                [CmdArgType.DevDouble, AttrDataFormat.SCALAR, AttrWriteType.READ],
                {"display level": "unknown"},
            ],
            "Wrong display level",
        ),
        (
            [
                [CmdArgType.DevDouble, AttrDataFormat.SCALAR, AttrWriteType.READ],
                {"polling period": "unknown"},
            ],
            "Wrong polling period",
        ),
        (
            [
                [CmdArgType.DevDouble, AttrDataFormat.SCALAR, AttrWriteType.READ],
                {"memorized": "unknown"},
            ],
            "Wrong memorized value",
        ),
        (
            [[CmdArgType.DevEnum, AttrDataFormat.SCALAR, AttrWriteType.READ], {}],
            "Missing 'enum_labels' key in attr_list definition",
        ),
    ],
)
def test_from_attr_info_exceptions(attr_info, exception_message):

    attr_name = "some_attr"
    class_name = "some_class"

    with pytest.raises(DevFailed, match=exception_message):
        AttrData(attr_name, class_name, attr_info=attr_info)


def test_from_attr_info_hw_memorized():

    attr_name = "some_attr"
    class_name = "some_class"

    attr_info = [
        [CmdArgType.DevDouble, AttrDataFormat.SCALAR, AttrWriteType.READ],
        {"memorized": "TRUE"},
    ]
    attr_data = AttrData(attr_name, class_name, attr_info=attr_info)
    assert attr_data.memorized
    assert attr_data.hw_memorized


def test_from_attr_info_memorized():

    attr_name = "some_attr"
    class_name = "some_class"

    attr_info = [
        [CmdArgType.DevDouble, AttrDataFormat.SCALAR, AttrWriteType.READ],
        {"memorized": "true_without_hard_applied"},
    ]
    attr_data = AttrData(attr_name, class_name, attr_info=attr_info)
    assert attr_data.memorized
    assert not attr_data.hw_memorized


# The following devices and fixture used to force rare segfault, when Device used as fixture
class BaseFixtureDevice(Device):
    def init_device(self):
        self.driver = self.create_driver()

    @command
    def cmd(self):
        raise RuntimeError("Bad command")

    def update_callback(self, data):
        pass


@pytest.fixture
def mocked_driver():
    return {}


@pytest.fixture
def TestFixtureDevice(mocked_driver):
    class D(BaseFixtureDevice):
        def create_driver(self):
            mocked_driver["callback"] = self.update_callback
            return mocked_driver

    return D


def test_device_repr_does_not_segfault_with_pytest(mocked_driver, TestFixtureDevice):
    with DeviceTestContext(TestFixtureDevice) as dp:
        with pytest.raises(DevFailed):
            dp.cmd()


def test_client_ident(server_green_mode):

    class MyDevice(Device):

        green_mode = server_green_mode

        if server_green_mode == GreenMode.Asyncio:

            @command()
            async def test_client_ident(self):
                self.assert_client_ident_is_none_for_non_sync_servers()

        elif server_green_mode == GreenMode.Gevent:

            @command()
            def test_client_ident(self):
                self.assert_client_ident_is_none_for_non_sync_servers()

        else:

            @command()
            def test_client_ident(self):
                self.assert_client_ident_valid()

        def assert_client_ident_valid(self):
            data = self.get_client_ident()
            assert data.client_ident
            assert data.client_lang == LockerLanguage.CPP_6
            assert data.client_pid == os.getpid()
            assert data.client_ip == "collocated client (c++ to c++ call)"
            data2 = self.get_client_ident()
            assert data2 == data

        def assert_client_ident_is_none_for_non_sync_servers(self):
            assert self.get_client_ident() is None

    with DeviceTestContext(MyDevice) as dev:
        dev.test_client_ident()


@pytest.mark.asyncio
async def test___to_callback():

    def callback1():
        return "called"

    def callback2(a, b):
        return f"called {a=}, {b=}"

    def callback3(a, b, c=0):
        return f"called {a=}, {b=}, {c=}"

    cb = __to_callback(None, "None-becomes-callable", GreenMode.Synchronous)
    assert callable(cb)

    cb = __to_callback(callback1, "callable", GreenMode.Synchronous)
    assert callable(cb)
    assert cb() == "called"

    cb = __to_callback([callback1], "seq-length-1", GreenMode.Synchronous)
    assert callable(cb)
    assert cb() == "called"

    args = (1, 2)
    cb = __to_callback([callback2, args], "seq-length-2", GreenMode.Synchronous)
    assert callable(cb)
    assert cb() == "called a=1, b=2"

    kwargs = {"c": 3}
    cb = __to_callback([callback3, args, kwargs], "seq-length-3", GreenMode.Synchronous)
    assert callable(cb)
    assert cb() == "called a=1, b=2, c=3"

    with pytest.raises(TypeError):
        _ = __to_callback({}, "wrong-type", GreenMode.Synchronous)

    with pytest.raises(TypeError):
        _ = __to_callback([], "seq-length-0", GreenMode.Synchronous)

    with pytest.raises(TypeError):
        _ = __to_callback(["invalid"], "seq-length-1-bad-type", GreenMode.Synchronous)

    with pytest.raises(TypeError):
        _ = __to_callback([callback1, [], {}, 4], "seq-length-4", GreenMode.Synchronous)

    cb = __to_callback(callback1, "sync-func-becomes-coroutine", GreenMode.Asyncio)
    assert inspect.iscoroutinefunction(cb)
    result = await cb()
    assert result == "called"

    cb = __to_callback(asyncio.sleep, "coroutine-unchanged", GreenMode.Asyncio)
    assert cb is asyncio.sleep


def test_no_crash_when_error_in_delete_device(capfd):

    class TestDevice(Device):
        def delete_device(self):
            raise RuntimeError("Don't crash please")

    with DeviceTestContext(TestDevice) as dev:
        dev.ping()

    _, err = capfd.readouterr()

    assert "delete_device() raised a DevFailed exception" in err
    assert "RuntimeError: Don't crash please" in err


def test_polling_configuration():
    class TestDevice(Device):

        def init_device(self):
            # check polling from attribute/command definition
            assert self.is_attribute_polled("attr")
            assert self.get_attribute_poll_period("attr") == 1000
            assert self.is_command_polled("cmd")
            assert self.get_command_poll_period("cmd") == 2000

            # check polling can be changed
            self.poll_attribute("attr", 1100)
            assert self.is_attribute_polled("attr")
            assert self.get_attribute_poll_period("attr") == 1100
            self.poll_command("cmd", 2100)
            assert self.is_command_polled("cmd")
            assert self.get_command_poll_period("cmd") == 2100

            # check polling can be stopped
            self.stop_poll_attribute("attr")
            assert not self.is_attribute_polled("attr")
            self.stop_poll_command("cmd")
            assert not self.is_command_polled("cmd")

            # restart polling for client to check
            self.poll_attribute("attr", 1200)
            self.poll_command("cmd", 2200)

        @attribute(polling_period=1000)
        def attr(self) -> int:
            return 55

        @command(polling_period=2000)
        def cmd(self):
            pass

    with DeviceTestContext(TestDevice) as proxy:
        # check polling enabled
        assert proxy.is_attribute_polled("attr")
        assert proxy.get_attribute_poll_period("attr") == 1200
        assert proxy.is_command_polled("cmd")
        assert proxy.get_command_poll_period("cmd") == 2200
        # check polling can be stopped
        proxy.stop_poll_attribute("attr")
        assert not proxy.is_attribute_polled("attr")
        proxy.stop_poll_command("cmd")
        assert not proxy.is_command_polled("cmd")



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
from tango import AttrWriteType, DeviceProxy, InfoIt
from tango.server import Device, attribute, command, device_property


class Leader(Device):
    FollowerTRLs = device_property(dtype=(str,))

    @InfoIt(show_args=True)
    @command(dtype_in=int)
    def Turn_Follower_On(self, follower_id):
        follower_trl = self.FollowerTRLs[follower_id - 1]
        follower_device = DeviceProxy(follower_trl)
        follower_device.is_on = True

    @command(dtype_in=int)
    def Turn_Follower_Off(self, follower_id):
        follower_trl = self.FollowerTRLs[follower_id - 1]
        follower_device = DeviceProxy(follower_trl)
        follower_device.is_on = False


class Follower(Device):
    def init_device(self):
        super().init_device()
        self._is_on = False

    is_on = attribute(
        dtype=bool,
        access=AttrWriteType.READ_WRITE,
    )

    @InfoIt(show_ret=True)
    def read_is_on(self):
        return self._is_on

    @InfoIt(show_args=True)
    def write_is_on(self, value):
        self._is_on = value


devices_info = [
    {
        "class": Leader,
        "devices": (
            {
                "name": "device/leader/1",
                "properties": {
                    "FollowerTRLs": ["device/follower/1", "device/follower/2"],
                },
            },
        ),
    },
    {
        "class": Follower,
        "devices": [
            {"name": "device/follower/1", "properties": {}},
            {"name": "device/follower/2", "properties": {}},
        ],
    },
]


class TestLeaderFollowerIntegration:
    def test_leader_turn_follower_on(self, tango_context):
        leader = DeviceProxy("device/leader/1")
        follower_1 = DeviceProxy("device/follower/1")
        follower_2 = DeviceProxy("device/follower/2")

        # check initial state: both followers are off
        assert follower_1.is_on == False
        assert follower_2.is_on == False

        # tell leader to enable follower_1
        leader.turn_follower_on(1)

        # check follower_1 is now on, and follower_2 is still off
        assert follower_1.is_on == True
        assert follower_2.is_on == False

    def test_leader_turn_follower_off(self, tango_context):
        leader = DeviceProxy("device/leader/1")
        follower_1 = DeviceProxy("device/follower/1")
        follower_2 = DeviceProxy("device/follower/2")

        # tell leader to enable both followers
        leader.turn_follower_on(1)
        leader.turn_follower_on(2)

        # check initial state: both followers are on
        assert follower_1.is_on == True
        assert follower_2.is_on == True

        # tell leader to disable follower_1
        leader.turn_follower_off(1)

        # check follower_1 is now off, and follower_2 is still on
        assert follower_1.is_on == False
        assert follower_2.is_on == True



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""
This is an internal PyTango module.
"""

__all__ = ("Release",)

__docformat__ = "restructuredtext"


class Release:
    """Summarize release information as class attributes.

    Release information:
        - name: (str) package name
        - version_info: (tuple<int,int,int,str,int>) The five components
          of the version number: major, minor, micro, releaselevel, and
          serial.
        - version: (str) package version in format <major>.<minor>.<micro>
        - release: (str) pre-release, post-release or development release;
          it is empty for final releases.
        - version_long: (str) package version in format
          <major>.<minor>.<micro><releaselevel><serial>
        - version_description: (str) short description for the current
          version
        - version_number: (int) <major>*100 + <minor>*10 + <micro>
        - description : (str) package description
        - long_description: (str) longer package description
        - authors: (dict<str(last name), tuple<str(full name),str(email)>>)
          package authors
        - url: (str) package url
        - download_url: (str) package download url
        - platform: (seq<str>) list of available platforms
        - keywords: (seq<str>) list of keywords
        - license: (str) the license
    """

    name = "pytango"
    version_info = (10, 1, 4)
    version = ".".join(map(str, version_info[:3]))
    release = "".join(map(str, version_info[3:]))
    separator = "." if "dev" in release or "post" in release else ""
    version_long = version + separator + release

    version_description = "This version implements the C++ Tango 10.1 API."
    version_number = int(version.replace(".", ""))
    description = "A python binding for the Tango control system"
    long_description = "This module implements the Python Tango Device API mapping."
    license = "LGPL"
    authors = {"Coutinho": ("Tiago Coutinho", "coutinho@esrf.fr")}
    author_lines = "\n".join([f"{x[0]} <{x[1]}>" for x in authors.values()])
    url = "http://gitlab.com/tango-controls/pytango"
    download_url = "http://pypi.python.org/pypi/pytango"
    platform = ["Linux", "Windows", "macOS"]
    keywords = ["Tango", "CORBA", "binding"]



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""
This is an internal PyTango module. It provides tango log classes that can
be used as decorators in any method of :class:`tango.DeviceImpl`.

To access these members use directly :mod:`tango` module and NOT tango.log4tango.

Example::

    import tango

    class MyDev(tango.Device_4Impl):

        @tango.InfoIt()
        def read_Current(self, attr):
            attr.set_value(self._current)
"""

__all__ = ("TangoStream", "LogIt", "DebugIt", "InfoIt", "WarnIt", "ErrorIt", "FatalIt")

__docformat__ = "restructuredtext"

import functools

from tango.utils import _is_coroutine_function

MAX_ARG_LEN = 25


def _compact_elem(v):
    v = repr(v)
    if len(v) > MAX_ARG_LEN:
        v = v[: MAX_ARG_LEN - 5] + "[...]"
    return v


def _compact_elems(elems):
    return map(_compact_elem, elems)


def _compact_elems_str(elems):
    return ", ".join(_compact_elems(elems))


def _compact_dict_str(d):
    return ", ".join(f"{k}={_compact_elem(v)}" for k, v in d.items())


class TangoStream:
    def __init__(self, fn):
        self._fn = fn
        self._accum = ""

    def write(self, s):
        self._accum += s
        # while there is no new line, just accumulate the buffer
        try:
            if s[-1] == "\n" or s.index("\n") >= 0:
                self.flush()
        except ValueError:
            pass

    def flush(self):
        b = self._accum
        if b is None or len(self._accum) == 0:
            return
        # take the '\n' because the log adds it
        if b[-1] == "\n":
            b = b[:-1]
        self._fn(b)
        self._accum = ""


class LogIt:
    """A class designed to be a decorator of any method of a
    :class:`tango.DeviceImpl` subclass. The idea is to log the entrance and
    exit of any decorated method.

    Example::

        class MyDevice(tango.Device_4Impl):

            @tango.LogIt()
            def read_Current(self, attr):
                attr.set_value(self._current, 1)

    All log messages generated by this class have DEBUG level. If you whish
    to have different log level messages, you should implement subclasses that
    log to those levels. See, for example, :class:`tango.InfoIt`.

    The constructor receives three optional arguments:
        * show_args - shows method arguments in log message (defaults to False)
        * show_kwargs - shows keyword method arguments in log message (defaults to False)
        * show_ret - shows return value in log message (defaults to False)
    """

    def __init__(self, show_args=False, show_kwargs=False, show_ret=False):
        """Initializes de LogIt object.

        :param show_args: (bool) show arguments in log message (default is False)
        :param show_kwargs: (bool) show keyword arguments in log message (default is False)
        :param show_ret: (bool) show return in log message (default is False)
        """
        self._show_args = show_args
        self._show_kwargs = show_kwargs
        self._show_ret = show_ret

    def is_enabled(self, obj):
        return obj.get_logger().is_debug_enabled()

    def get_log_func(self, obj):
        return obj.debug_stream

    def __call__(self, f):
        if _is_coroutine_function(f):

            @functools.wraps(f)
            async def log_stream(*args, **kwargs):
                dev = args[0]
                if not self.is_enabled(dev):
                    return await f(*args, **kwargs)
                log = self.get_log_func(dev)
                f_name = dev.__class__.__name__ + "." + f.__name__
                f_args = []
                if self._show_args:
                    f_args.append(_compact_elems_str(args[1:]))
                if self._show_kwargs:
                    f_args.append(_compact_dict_str(kwargs))
                f_args_str = ", ".join(f_args)
                log(f"-> {f_name}({f_args_str})", source=f)
                with_exc = True
                try:
                    ret = await f(*args, **kwargs)
                    with_exc = False
                    return ret
                finally:
                    if with_exc:
                        log(f"<- {f_name}() raised exception!", source=f)
                    else:
                        sret = ""
                        if self._show_ret:
                            sret = _compact_elem(ret) + " "
                        log(f"{sret}<- {f_name}()", source=f)

        else:

            @functools.wraps(f)
            def log_stream(*args, **kwargs):
                dev = args[0]
                if not self.is_enabled(dev):
                    return f(*args, **kwargs)
                log = self.get_log_func(dev)
                f_name = dev.__class__.__name__ + "." + f.__name__
                f_args = []
                if self._show_args:
                    f_args.append(_compact_elems_str(args[1:]))
                if self._show_kwargs:
                    f_args.append(_compact_dict_str(kwargs))
                f_args_str = ", ".join(f_args)
                log(f"-> {f_name}({f_args_str})", source=f)
                with_exc = True
                try:
                    ret = f(*args, **kwargs)
                    with_exc = False
                    return ret
                finally:
                    if with_exc:
                        log(f"<- {f_name}() raised exception!", source=f)
                    else:
                        sret = ""
                        if self._show_ret:
                            sret = _compact_elem(ret) + " "
                        log(f"{sret}<- {f_name}()", source=f)

        log_stream._wrapped = f
        return log_stream


class DebugIt(LogIt):
    """A class designed to be a decorator of any method of a
    :class:`tango.DeviceImpl` subclass. The idea is to log the entrance and
    exit of any decorated method as DEBUG level records.

    Example::

        class MyDevice(tango.Device_4Impl):

            @tango.DebugIt()
            def read_Current(self, attr):
                attr.set_value(self._current, 1)

    All log messages generated by this class have DEBUG level.

    The constructor receives three optional arguments:
        * show_args - shows method arguments in log message (defaults to False)
        * show_kwargs - shows keyword method arguments in log message (defaults to False)
        * show_ret - shows return value in log message (defaults to False)
    """

    def is_enabled(self, d):
        return d.get_logger().is_debug_enabled()

    def get_log_func(self, d):
        return d.debug_stream


class InfoIt(LogIt):
    """A class designed to be a decorator of any method of a
    :class:`tango.DeviceImpl` subclass. The idea is to log the entrance and
    exit of any decorated method as INFO level records.

    Example::

        class MyDevice(tango.Device_4Impl):

            @tango.InfoIt()
            def read_Current(self, attr):
                attr.set_value(self._current, 1)

    All log messages generated by this class have INFO level.

    The constructor receives three optional arguments:
        * show_args - shows method arguments in log message (defaults to False)
        * show_kwargs - shows keyword method arguments in log message (defaults to False)
        * show_ret - shows return value in log message (defaults to False)
    """

    def is_enabled(self, d):
        return d.get_logger().is_info_enabled()

    def get_log_func(self, d):
        return d.info_stream


class WarnIt(LogIt):
    """A class designed to be a decorator of any method of a
    :class:`tango.DeviceImpl` subclass. The idea is to log the entrance and
    exit of any decorated method as WARN level records.

    Example::

        class MyDevice(tango.Device_4Impl):

            @tango.WarnIt()
            def read_Current(self, attr):
                attr.set_value(self._current, 1)

    All log messages generated by this class have WARN level.

    The constructor receives three optional arguments:
        * show_args - shows method arguments in log message (defaults to False)
        * show_kwargs - shows keyword method arguments in log message (defaults to False)
        * show_ret - shows return value in log message (defaults to False)
    """

    def is_enabled(self, d):
        return d.get_logger().is_warn_enabled()

    def get_log_func(self, d):
        return d.warn_stream


class ErrorIt(LogIt):
    """A class designed to be a decorator of any method of a
    :class:`tango.DeviceImpl` subclass. The idea is to log the entrance and
    exit of any decorated method as ERROR level records.

    Example::

        class MyDevice(tango.Device_4Impl):

            @tango.ErrorIt()
            def read_Current(self, attr):
                attr.set_value(self._current, 1)

    All log messages generated by this class have ERROR level.

    The constructor receives three optional arguments:
        * show_args - shows method arguments in log message (defaults to False)
        * show_kwargs - shows keyword method arguments in log message (defaults to False)
        * show_ret - shows return value in log message (defaults to False)
    """

    def is_enabled(self, d):
        return d.get_logger().is_error_enabled()

    def get_log_func(self, d):
        return d.error_stream


class FatalIt(LogIt):
    """A class designed to be a decorator of any method of a
    :class:`tango.DeviceImpl` subclass. The idea is to log the entrance and
    exit of any decorated method as FATAL level records.

    Example::

        class MyDevice(tango.Device_4Impl):

            @tango.FatalIt()
            def read_Current(self, attr):
                attr.set_value(self._current, 1)

    All log messages generated by this class have FATAL level.

    The constructor receives three optional arguments:
        * show_args - shows method arguments in log message (defaults to False)
        * show_kwargs - shows keyword method arguments in log message (defaults to False)
        * show_ret - shows return value in log message (defaults to False)
    """

    def is_enabled(self, d):
        return d.get_logger().is_fatal_enabled()

    def get_log_func(self, d):
        return d.fatal_stream
